/******************************** Module Header *********************************
* Program Name: snoop.c
*
* ZoomIn utility.  This tool magnifies a portion of the screen,
* allowing you to see things at a pixel level.
*
* Based on Microsoft Zoomin utility source extended by Thierry Maurel 
* (tmaurel@caramail.com) found at http://codeguru.earthweb.com.
*
* Modified by Yung-Yu Chuang (cyy@cs.washington.edu) 
*
* History:
* 09/05/2000	Created. 
*
* 09/10/2000	Ver 0.01: 
*					remove vertical scroll bar
*					9 zoom factors(1,2,3,4,6,8,10,16,32) changed by + and -
*					add status bar to display coordinate and RGB	
*					add RGB intensity graph 
* 09/15/2000	Ver 0.1:
*					add arrow mode
*					add 'd' to display the capturing region
*					add the capability to show coordinate wrt the active window
*					code cleanup
*
* 03/18/2001	Ver 0.22:
*					add Hex display mode
*
*********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "snoop.h"
#include "ezfont.h"
#include "registry.h"

#include "globals.h"

// Brian's favor
#define BRIAN		1

// about zoom factors
#define MIN_ZOOM_INDEX		0		// it should be always 0
#define MAX_ZOOM_INDEX		8		// the number of zoom factor minus 1

CHAR szAppName[] = "Snoop";				// Aplication name.
HINSTANCE ghInst;						// Instance handle.
HWND ghwndApp;							// Main window handle.
HWND ghwndStatusBar;					// Status bar handle.
HACCEL ghaccelTable;					// Main accelerator table.
HPALETTE ghpalPhysical;				    // Handle to the physical palette.

// pens 
static HPEN ghRPen, ghGPen, ghBPen, ghR2Pen;
static HPEN ghGrayPen;
static HBRUSH ghGrayBrush;

HFONT ghPeakFont;

INT gnDisplayMode = MENU_FILTER_SNOOP;

float screenScaleFactorY;	//extern in globals.h

/************************************************************************
* WinMain
*
* Main entry point for the application.
*
* Arguments:
*
* History:
*
************************************************************************/

INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, INT nCmdShow)
{
	
	initScreenScaleFactor();// do this BEFORE SetProcessDPIAware()

	SetProcessDPIAware();

    MSG msg;
	// copy command options 
	strcpy(gszCmdLine, lpCmdLine);

    if (!InitInstance(hInst, nCmdShow))
        return FALSE;

	

    /*     
	 * Polling messages from event queue
     */

    while (GetMessage(&msg, NULL, 0, 0)) {
        if (!TranslateAccelerator(ghwndApp, (HACCEL)ghaccelTable, &msg)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    return msg.wParam;
}



/************************************************************************
* InitInstance
*
* Instance initialization for the app.
*
* Arguments:
*
* History:
*
************************************************************************/

BOOL InitInstance(
    HINSTANCE hInst,
    INT cmdShow)
{
    

	WNDCLASS wc;
    INT dx;
    INT dy;
    DWORD flStyle;
    RECT rc;

    ghInst = hInst;

    /*
     * Register a class for the main application window.
     */
    wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
    wc.hIcon          = LoadIcon(hInst, "snoop");
    wc.lpszMenuName   = MAKEINTRESOURCE(IDMENU_SNOOP);
    wc.lpszClassName  = szAppName;
    wc.hbrBackground  = (HBRUSH)GetStockObject(BLACK_BRUSH);
    wc.hInstance      = hInst;
    wc.style          = CS_BYTEALIGNCLIENT | CS_VREDRAW | CS_HREDRAW;
    wc.lpfnWndProc    = (WNDPROC)AppWndProc;
    wc.cbWndExtra     = 0;
    wc.cbClsExtra     = 0;

    if (!RegisterClass(&wc))
        return FALSE;

    if (!(ghaccelTable = LoadAccelerators(hInst, MAKEINTRESOURCE(IDACCEL_SNOOP))))
        return FALSE;

    if (!(ghpalPhysical = CreatePhysicalPalette()))
        return FALSE;

    gcxScreenMax = GetSystemMetrics(SM_CXSCREEN) - 1;    gcyScreenMax = GetSystemMetrics(SM_CYSCREEN) - 1;

//	gnColorBits = GetSystemMetrics(BITSPIXEL);

    flStyle = WS_CAPTION | WS_OVERLAPPED | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX;

	gnZoom = gnZoomFactors[gnZoomIndex];
#ifdef BRIAN
    dx = 300;
    dy = 300+ SB_HEIGHT;
#else
    dx = 20 * gnZoom;
    dy = 20 * gnZoom + SB_HEIGHT;
#endif

    SetRect(&rc, 0, 0, dx, dy);
    AdjustWindowRect(&rc, flStyle, TRUE);

#ifdef BRIAN

    ghwndApp = CreateWindowEx(
		WS_EX_TOPMOST, 
		szAppName, 
		szAppName, 
		flStyle,
        300, 
		300, 
		(rc.right - rc.left), 
		(rc.bottom - rc.top),
		NULL, 
		NULL, 
		hInst, 
		NULL);
#else
    ghwndApp = CreateWindowEx(WS_EX_TOPMOST, szAppName, szAppName, flStyle,
            gcxScreenMax - dx - 10, gcyScreenMax - dy - 75, rc.right - rc.left, rc.bottom - rc.top,
            NULL, NULL, hInst, NULL);
#endif

    if (!ghwndApp)
        return FALSE;

	// Snoop specific
	// allocate buffer for graph drawing
	gcolorBuf = (COLORREF *)malloc(gcxScreenMax*sizeof(COLORREF));
	gptBuf[0] = (POINT *)malloc(gcxScreenMax*sizeof(POINT));
	gptBuf[1] = (POINT *)malloc(gcxScreenMax*sizeof(POINT));
	gptBuf[2] = (POINT *)malloc(gcxScreenMax*sizeof(POINT));

	// set init timer
	if (gfRefEnable) {
		if (!SetTimer(ghwndApp, IDTIMER_SNOOP, gnRefInterval * 10, NULL))
			gfRefEnable = FALSE;
	}
	// end Snoop specific

    ShowWindow(ghwndApp, cmdShow);
	

	ResetTitle();

    return TRUE;
}



/************************************************************************
* CreatePhysicalPalette
*
* Creates a palette for the app to use.  The palette references the
* physical palette, so that it can properly display images grabbed
* from palette managed apps.
*
* History:
*
************************************************************************/

HPALETTE CreatePhysicalPalette(VOID)
{
    PLOGPALETTE ppal;
    HPALETTE hpal = NULL;
    INT i;

    ppal = (PLOGPALETTE)LocalAlloc(LPTR,
            sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * NPAL);
    if (ppal) {
        ppal->palVersion = 0x300;
        ppal->palNumEntries = NPAL;

        for (i = 0; i < NPAL; i++) {
            ppal->palPalEntry[i].peFlags = (BYTE)PC_EXPLICIT;
            ppal->palPalEntry[i].peRed   = (BYTE)i;
            ppal->palPalEntry[i].peGreen = (BYTE)0;
            ppal->palPalEntry[i].peBlue  = (BYTE)0;
        }

        hpal = CreatePalette(ppal);
        LocalFree(ppal);
    }

    return hpal;
}


int findIndex(int zoom)
{
	int i;

	for (i=MAX_ZOOM_INDEX; i>=MIN_ZOOM_INDEX; i--) {
		if (zoom>=gnZoomFactors[i]) return i;
	}

	return 0;
}

void ProcessCommandLine(char *gszCmdLine, HWND hwnd, HMENU hmenu)
{
	char *token;
	int x, y, w, h;
	BOOL bSetPos=FALSE;
	int conf;

	token=strtok(gszCmdLine, " -");
	while (token!=NULL) {
		if (strcmp(token, "top")==0) {
			gbOnTop = TRUE;
			SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST), 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
			CheckMenuItem(hmenu, MENU_OPTIONS_ONTOP, (gbOnTop ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "notop")==0) {
			gbOnTop = FALSE;
			SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST), 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
			CheckMenuItem(hmenu, MENU_OPTIONS_ONTOP, (gbOnTop ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "grid")==0) {
			gbGrid = TRUE;
			CheckMenuItem(hmenu, MENU_GRID_ENABLED, (gbGrid ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "nogrid")==0) {
			gbGrid = FALSE;
			CheckMenuItem(hmenu, MENU_GRID_ENABLED, (gbGrid ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "ruler")==0) {
			gbRuler = TRUE;
			CheckMenuItem(hmenu, MENU_RULER_ENABLED, (gbRuler ? MF_CHECKED : MF_UNCHECKED));
			// recalculate the size
			CalcZoomedSize();
			CalcRGBGraphSize();
		} else if (strcmp(token, "noruler")==0) {
			gbRuler = FALSE;
			CheckMenuItem(hmenu, MENU_RULER_ENABLED, (gbRuler ? MF_CHECKED : MF_UNCHECKED));
			// recalculate the size
			CalcZoomedSize();
			CalcRGBGraphSize();
		} else if (strcmp(token, "highlight")==0) {
			gbRGBGraphHighlight = TRUE;
			CheckMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, (gbRGBGraphHighlight ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "nohighlight")==0) {
			gbRGBGraphHighlight = FALSE;
			CheckMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, (gbRGBGraphHighlight ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "centermark")==0) {
            gbCenterHighlight = TRUE;
			CheckMenuItem(hmenu, MENU_CENTER_HIGHLIGHT_ENABLED, (gbCenterHighlight ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "nocentermark")==0) {
            gbCenterHighlight = FALSE;
			CheckMenuItem(hmenu, MENU_CENTER_HIGHLIGHT_ENABLED, (gbCenterHighlight ? MF_CHECKED : MF_UNCHECKED));
		} else if (strcmp(token, "shift")==0) {
			token=strtok(NULL, " -");
			gnFastDelta=atoi(token);
		} else if (strcmp(token, "f")==0) {
			token=strtok(NULL, " -");
			conf=atoi(token);
			LoadRegistry(hwnd, hmenu, conf);
		} else if (strcmp(token, "coord")==0) {
			token=strtok(NULL, " -");
			x=atoi(token);
			token=strtok(NULL, " -");
			y=atoi(token);					

			int w, h;
			GetWindowsDimension(&w, &h);

			if (x<w && y<h)
				SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST),x,y,0,0, SWP_NOSIZE );
			else
				SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST),0,0,0,0, SWP_NOSIZE );

			bSetPos = TRUE;
		} else if (strcmp(token, "size")==0) {
			token=strtok(NULL, " -");
			w=atoi(token);
			token=strtok(NULL, " -");
			h=atoi(token);

			h=h+54;	// title+menu+status bar

			if (bSetPos)
				SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST),0,0,w,h, SWP_NOMOVE );
			else {
				x = gcxScreenMax - w - 2;
				y = gcyScreenMax - h - 28;
				SetWindowPos(hwnd, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST),x,y,w,h, SWP_DRAWFRAME);
			}
		} else if (strcmp(token, "inversey")==0) {
			gbInverseY=TRUE;
		} else if (strcmp(token, "hex")==0) {
			gbDisplayHex=TRUE;
		} else if (strcmp(token, "bound")==0) {
			gbBound=TRUE;
		} else {
			SendMessage(hwnd, WM_CHAR, token[0], 0L);
		}
		token=strtok(NULL, " -");
	}

	gbInit = FALSE;
}

int DetectColorDepth()
{
	BITMAP bitmap;
	HBITMAP hBitmap;
	HDC hdc;

	hdc = GetDC(ghwndApp);
	hBitmap = CreateCompatibleBitmap(hdc, 4, 4);
	
	GetObject(hBitmap, sizeof(BITMAP), (LPVOID)&bitmap);

	DeleteObject(hBitmap);
	ReleaseDC(ghwndApp, hdc);

	return bitmap.bmBitsPixel;
}

void GetWindowsDimension(int *width, int *height)
{
	DEVMODE devmode;
	
	devmode.dmSize=sizeof(DEVMODE);
	devmode.dmDriverExtra=0;

	EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &devmode);
	*width  = devmode.dmPelsWidth;
	*height = devmode.dmPelsHeight;
}

/***********************************************************************
 *
 *  For color space adjustment
 *
 ***********************************************************************/
VOID FillAdjustmentLUT(unsigned char *lut, int x1, int y1, int x2, int y2)
{
	int n, i, ii, x;

	n=x2-x1;

	if (n==0) return;

	i =(y1 << 16);
	ii=( (y2-y1+1) << 16)/n;

	for (x=x1; x<=x2; x++) {
		lut[x]=i>>16;
		i+=ii;
	}
}

int gnBrightness=0;
int gnContrast=0;
float gfGamma=1.0f;

unsigned char BrightnessLUT[256];
unsigned char ContrastLUT[256];
unsigned char GammaLUT[256];

VOID SetBrightnessLUT(int amount)
{
	amount=(int)(amount*255.0/100.0+0.5);

	if (amount<0)
		FillAdjustmentLUT(BrightnessLUT, 0, 0, 255, 255+amount);
	else {
		if (amount>255) amount=255;
		FillAdjustmentLUT(BrightnessLUT, 0, amount, 255, 255);
	}
}

VOID SetContrastLUT(int amount)
{
	int x;
	
	// remapping from UI to real amount
	amount=(int)(amount*128.0/100.0+0.5);

	if (amount<0)
		FillAdjustmentLUT(ContrastLUT, 0, -amount, 255, 255+amount);
	else {
		if (amount>255) amount=255;
		for (x=0; x<amount; x++) ContrastLUT[x]=0;
		FillAdjustmentLUT(ContrastLUT, amount, 0, 255-amount, 255);
		for (x=255; x>255-amount; x--) ContrastLUT[x]=255;
	}
}

VOID SetGammaLUT(float amount)
{
	float y;
	int z, i;

	if (amount>0) {
		GammaLUT[0]=0;
		y=1/amount;
		for (i=1; i<256; i++) {
			z=(int)(255*exp(y*log(i/255.0))+0.5);
			if (z>255) z=255;
			else if (z<0) z=0;
			GammaLUT[i]=z;
		}
	}
}

VOID AdjustColor(unsigned char *buf, int width, int height, int padding)
{
	int x, y;
	unsigned char *pData;

	pData=buf;
	for (y=0; y<height; y++) {
		for (x=0; x<width; x++) {
			pData[0]=ContrastLUT[BrightnessLUT[GammaLUT[pData[0]]]];
			pData[1]=ContrastLUT[BrightnessLUT[GammaLUT[pData[1]]]];
			pData[2]=ContrastLUT[BrightnessLUT[GammaLUT[pData[2]]]];
			if (gnColorBits==32) pData+=4;
			else pData+=3;
		}
		pData+=padding;
	}
}

/***********************************************************************
 *
 *  For screen capture.
 *
 ***********************************************************************/
// private, only used for capture routines
BITMAPINFO bmiCapture;
BITMAP bitmapCapture;
HBITMAP hBmpCaptureMem, hBmpCaptureMemOld;
HDC hdcCaptureMem;

// global, accessed by the others
int gnCaptureMemoryWidth;
int gnCaptureWidth, gnCaptureHeight;
int gnCapturePadding;

unsigned char *gpCaptureBuf;

VOID InitScreenCapture()
{
	bmiCapture.bmiHeader.biSize=sizeof(bmiCapture.bmiHeader);
	bmiCapture.bmiHeader.biPlanes=1;
	bmiCapture.bmiHeader.biBitCount=gnColorBits;
	bmiCapture.bmiHeader.biCompression=BI_RGB;
	bmiCapture.bmiHeader.biSizeImage=0;
	bmiCapture.bmiHeader.biClrImportant=0;
	bmiCapture.bmiHeader.biClrUsed=0;
}

VOID ScreenCapture(HDC hdcScreen, int x, int y, int width, int height, 
				   unsigned char **buf, int *padding)
{
	bmiCapture.bmiHeader.biWidth=width;
	bmiCapture.bmiHeader.biHeight=height;

    hdcCaptureMem = CreateCompatibleDC(hdcScreen);
	hBmpCaptureMem = CreateDIBSection(hdcCaptureMem, &bmiCapture, DIB_RGB_COLORS, (void **)buf, NULL, 0);
    hBmpCaptureMemOld = (HBITMAP)SelectObject(hdcCaptureMem, hBmpCaptureMem);

	// TODO: can we read it directly from screen's bitmap?
    // copy to the off-screen dc's bitmap
    BitBlt(hdcCaptureMem, 0, 0, width, height, hdcScreen, x, y, SRCCOPY);

	// get bitmap	
	GetObject(hBmpCaptureMem, sizeof(BITMAP), (LPVOID)&bitmapCapture);
	gnCaptureMemoryWidth=bitmapCapture.bmWidthBytes;
	gnCaptureHeight=bitmapCapture.bmHeight;
	gnCaptureWidth=bitmapCapture.bmWidth;
	gnCapturePadding=gnCaptureMemoryWidth-gnCaptureWidth*(gnColorBits>>3);

	*padding = gnCapturePadding;

	gpCaptureBuf = *buf;
}

VOID ReleaseScreenCapture()
{
    SelectObject(hdcCaptureMem, hBmpCaptureMemOld);
    DeleteObject(hBmpCaptureMem);
    DeleteDC(hdcCaptureMem);
}

/************************************************************************
* AppWndProc
*
* Main window proc for the zoomin utility.
*
* Arguments:
*   Standard window proc args.
*
* History:
*
************************************************************************/

LONG APIENTRY AppWndProc(
    HWND hwnd,
    UINT msg,
    WPARAM wParam,
    LPARAM lParam)
{
    PAINTSTRUCT ps;
	HMENU hmenu;
	
	LOGBRUSH lb;
	HDC hdc;

	hmenu = GetMenu(hwnd);

    switch (msg) {
        case WM_CREATE:

			// enable to use status bar 
			InitCommonControls();
			ghwndStatusBar = InitStatusBar(hwnd);

			// create pens to use
			ghRPen = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));
			ghGPen = CreatePen(PS_SOLID, 1, RGB(0, 255, 0));
			ghBPen = CreatePen(PS_SOLID, 1, RGB(0, 0, 255));

			ghR2Pen = CreatePen(PS_SOLID, 1, RGB(255, 0, 0));

			ghGrayPen = CreatePen(PS_SOLID, 1, RGB(192, 192, 192));
			lb.lbStyle=BS_SOLID;	lb.lbColor=RGB(192, 192, 192);
			ghGrayBrush = CreateBrushIndirect(&lb);

			hdc=GetDC(hwnd);
			ghPeakFont = EzCreateFont(hdc, "Courier New", 76, 76, EZ_ATTR_BOLD, TRUE);
			ReleaseDC(hwnd, hdc);

			// reflect init setup onto UI

			// check the current filter mode
			CheckMenuItem(hmenu, gnDisplayMode, MF_CHECKED);

			// check the current RGB graph mode
			CheckMenuItem(hmenu, gnRGBGraphMode, MF_CHECKED);
			CheckMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, (gbRGBGraphHighlight ? MF_CHECKED : MF_UNCHECKED));
			CheckMenuItem(hmenu, MENU_STATISTICS_INFO, (gbStatisticsInfo ? MF_CHECKED : MF_UNCHECKED));
			
			// default setup
			SendMessage(hwnd, WM_COMMAND, MENU_SCREEN, 0L);
			SendMessage(hwnd, WM_COMMAND, MENU_MOUSE_MODE, 0L);
			SendMessage(hwnd, WM_COMMAND, MENU_OPTIONS_ONTOP, 0L);
            SendMessage(hwnd, WM_COMMAND, MENU_CENTER_HIGHLIGHT_ENABLED, 0L);

			// load from registry
			LoadRegistry(hwnd, hmenu);

			CheckMenuItem(hmenu, MENU_EDIT_AUTO_SAVE_CONFIG, (gbSaveConfigExit ? MF_CHECKED : MF_UNCHECKED));

			// set up options according to command line
			ProcessCommandLine(gszCmdLine, hwnd, hmenu);

			if (gbInverseY)
				CheckMenuItem(hmenu, MENU_INVERSE_Y, MF_CHECKED);

			if (gbDisplayHex)
				CheckMenuItem(hmenu, MENU_DISPLAY_HEX, MF_CHECKED);

			gnColorBits = DetectColorDepth();

			GetWindowsDimension(&gnWindowsWidth, &gnWindowsHeight);

			InitScreenCapture();

			// set the color adjustment table
			SetBrightnessLUT(0);
			SetContrastLUT(0);
			SetGammaLUT(1.0f);

			ResetTitle();

			// disable menu items if necessary
			if (gnColorBits<24) {
				EnableMenuItem(hmenu, MENU_GRAPH_STATISTICS, MF_GRAYED);
				EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_GRAYED);
				EnableMenuItem(hmenu, MENU_FILTER_GRADIENT, MF_GRAYED);
				EnableMenuItem(hmenu, MENU_OPTIONS_HISTOGRAM, MF_GRAYED);
			}

			switch (gnRGBGraphMode) {
				case MENU_GRAPH_OFF:
					EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_GRAYED);
					EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_GRAYED);
					break;

				case MENU_GRAPH_HOR_STEP:
				case MENU_GRAPH_HOR_LINEAR:
				case MENU_GRAPH_VER_STEP:
				case MENU_GRAPH_VER_LINEAR:
					EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_ENABLED);
					EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_GRAYED);
					break;

				case MENU_GRAPH_STATISTICS:
					EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_GRAYED);
					EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_ENABLED);
					break;
			}

            break;

        case WM_TIMER:
            /*
             * Update on every timer message.  The cursor will be
             * flashed to the hourglash for some visual feedback
             * of when a snapshot is being taken.
             */

			// use timer to update window! No hook. Ya!
			if (gnMode==MENU_MOUSE_MODE) 
				GetCursorPos(&gptZoom);

            DoTheZoomIn(NULL);

            break;

        case WM_PAINT:
            BeginPaint(hwnd, &ps);
            DoTheZoomIn(ps.hdc);
            EndPaint(hwnd, &ps);
            return 0L;

        case WM_ERASEBKGND:
            return TRUE;
            break;

        case WM_SIZE:
			
			gnClientWidth  = LOWORD(lParam);
			gnClientHeight = HIWORD(lParam);

            CalcZoomedSize();

			CalcRGBGraphSize();

			// adjust atatus bar when window is resized
			SendMessage(ghwndStatusBar, WM_SIZE, wParam, lParam);

            break;

        case WM_LBUTTONDOWN:
            ((gptZoom).x = (SHORT)LOWORD(lParam), (gptZoom).y = (SHORT)HIWORD(lParam));
            ClientToScreen(hwnd, &gptZoom);
            DrawZoomRect();
            DoTheZoomIn(NULL);

            SetCapture(hwnd);
			gnOldMode = gnMode;

			gnMode = MENU_TRACK_MODE;

			ResetTitle();

            break;

        case WM_MOUSEMOVE:
            if (gnMode == MENU_TRACK_MODE) {
                DrawZoomRect();
                ((gptZoom).x = (SHORT)LOWORD(lParam), (gptZoom).y = (SHORT)HIWORD(lParam));
                ClientToScreen(hwnd, &gptZoom);
                DrawZoomRect();
                DoTheZoomIn(NULL);
            }

            break;

        case WM_LBUTTONUP:
            if (gnMode == MENU_TRACK_MODE) {
                DrawZoomRect();
                ReleaseCapture();

				if (gnMode==MENU_TRACK_MODE) {
					gnMode = gnOldMode;
				}

				ResetTitle();
            }

            break;

        case WM_KEYDOWN:
            switch (wParam) {
                case VK_UP:
                case VK_DOWN:
                case VK_LEFT:
                case VK_RIGHT:
					if ((GetKeyState(VK_SPACE) & 0x8000)){
						ResizeWindow((INT)wParam, GetKeyState(VK_SHIFT) & 0x8000);
						DrawZoomRect();
						Sleep(50);
						DrawZoomRect();
					} else {
						MoveView((INT)wParam, GetKeyState(VK_SHIFT) & 0x8000,
							    GetKeyState(VK_CONTROL) & 0x8000);
						DrawZoomRect();
						Sleep(50);
						DrawZoomRect();

						if (gnMode==MENU_MOUSE_MODE) {
							SetCursorPos(gptZoom.x, gptZoom.y);
						}
					} 
                    break;
				case VK_PRIOR:
					if ((GetKeyState(VK_SHIFT) & 0x8000)) {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// gamma
							gfGamma+=0.02f;
							if (gfGamma>2.0) gfGamma=2.0;
							SetGammaLUT(gfGamma);
						} else {
							// brightness
							gnBrightness+=2;
							if (gnBrightness>100) gnBrightness=100;
							SetBrightnessLUT(gnBrightness);
						}
					} else {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// contrast
							gnContrast+=2;
							if (gnContrast>100) gnContrast=100;
							SetContrastLUT(gnContrast);
						} else {
							// for histogram
							if (gnRGBGraphMode==MENU_GRAPH_STATISTICS) {
								gdScaleFactor-=0.1f;
								if (gdScaleFactor<0.05) gdScaleFactor=0.1f;
							}
						}
					}
/*
					if (gnDisplayMode==MENU_FILTER_SNOOP) {
						gdScaleFactor-=0.1f;
						if (gdScaleFactor<0.05) gdScaleFactor=0.1f;
					} else {
						gdGradientScaleFactor/=1.15f;
					}
*/
					break;
				case VK_NEXT:
					if ((GetKeyState(VK_SHIFT) & 0x8000)) {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// gamma
							gfGamma-=0.02f;
							if (gfGamma<0.02) gfGamma=0.02f;
							SetGammaLUT(gfGamma);
						} else {
							// brightness
							gnBrightness-=2;
							if (gnBrightness<-100) gnBrightness=-100;
							SetBrightnessLUT(gnBrightness);
						}
					} else {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// contrast
							gnContrast-=2;
							if (gnContrast<-100) gnContrast=-100;
							SetContrastLUT(gnContrast);
						} else {
							// for histogram
							if (gnRGBGraphMode==MENU_GRAPH_STATISTICS) {
								gdScaleFactor+=0.1f;
							}
						}
					}
/*
					if (gnDisplayMode==MENU_FILTER_SNOOP) {
						gdScaleFactor+=0.1f;
					} else {
						gdGradientScaleFactor*=1.15f;
					}
*/
					break;
				case VK_HOME:
					if ((GetKeyState(VK_SHIFT) & 0x8000)) {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// gamma
							gfGamma=1.0f;
							SetGammaLUT(gfGamma);
						} else {
							// brightness
							gnBrightness=0;
							SetBrightnessLUT(gnBrightness);
						}
					} else {
						if ((GetKeyState(VK_CONTROL) & 0x8000)) {
							// contrast
							gnContrast=0;
							SetContrastLUT(gnContrast);
						} else {
							// for histogram
							if (gnRGBGraphMode==MENU_GRAPH_STATISTICS) {
								gdScaleFactor=1.0f;
							}
						}
					}
/*
					if (gnDisplayMode==MENU_FILTER_SNOOP) {
						gdScaleFactor=1.0f;
					} else {
						gdGradientScaleFactor=1.0f;
					}
*/
					break;
	
				case VK_TAB:
					if (gnColorBits<24) break;
					CheckMenuItem(hmenu, gnDisplayMode, MF_UNCHECKED);
					gnDisplayMode++;
					if (gnDisplayMode>MENU_FILTER_LAST)
						gnDisplayMode=MENU_FILTER_FIRST;
					CheckMenuItem(hmenu, gnDisplayMode, MF_CHECKED);
					DoTheZoomIn(NULL);
					break;
/*				case VK_BACK:
					CheckMenuItem(hmenu, gnDisplayMode, MF_UNCHECKED);
					gnDisplayMode=MENU_FILTER_SNOOP;
					CheckMenuItem(hmenu, gnDisplayMode, MF_CHECKED);
					DoTheZoomIn(NULL);
					break;
*/
				case 68:	// 'd'
					DrawZoomRect();
					Sleep(200);
					DrawZoomRect();
					break;
                case VK_ESCAPE:
					SendMessage(hwnd, WM_COMMAND, MENU_EDIT_QUIT, 0L);
					break;
				case VK_F1:
					if (GetKeyState(VK_SHIFT) & 0x8000) {
						SaveRegistry(hwnd, 1);
					} else if (GetKeyState(VK_CONTROL) & 0x8000) {
						LoadRegistry(hwnd, hmenu, 1);
					} else {
						SendMessage(hwnd, WM_COMMAND, MENU_HELP_ABOUT, 0L);
					}
					break;
				case VK_F2:
					if (GetKeyState(VK_SHIFT) & 0x8000) {
						SaveRegistry(hwnd, 2);
					} else if (GetKeyState(VK_CONTROL) & 0x8000) {
						LoadRegistry(hwnd, hmenu, 2);
					} else {
						SendMessage(hwnd, WM_COMMAND, MENU_OPTIONS_ONTOP, 0L);
					}
					break;
				case VK_F3:
					if (GetKeyState(VK_SHIFT) & 0x8000) {
						SaveRegistry(hwnd, 3);
					} else if (GetKeyState(VK_CONTROL) & 0x8000) {
						LoadRegistry(hwnd, hmenu, 3);
					} else {
						SendMessage(hwnd, WM_COMMAND, MENU_GRID_ENABLED, 0L);
					}
					break;
				case VK_F4:
					if (GetKeyState(VK_SHIFT) & 0x8000) {
						SaveRegistry(hwnd, 4);
					} else if (GetKeyState(VK_CONTROL) & 0x8000) {
						LoadRegistry(hwnd, hmenu, 4);
					} else {
						if (gnColorBits<24) break;
						SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_STATISTICS, 0L);
					}
					break;
				case VK_F5:
				case VK_F6:
				case VK_F7:
				case VK_F8:
				case VK_F9:					
					if (GetKeyState(VK_SHIFT) & 0x8000) {
						SaveRegistry(hwnd, wParam-VK_F1+1);
					} else if (GetKeyState(VK_CONTROL) & 0x8000) {
						LoadRegistry(hwnd, hmenu, wParam-VK_F1+1);
					} else {
					}
					break;
				case VK_BACK:
					LoadRegistry(hwnd, hmenu, NUM_CONFIG);
					SaveRegistry(hwnd, 1);
					break;
				case VK_DELETE:
					// remove all registry
					RemoveRegistry();
					break;
            }

            break;

		case WM_CHAR:
			switch (wParam) {
				case '+':
				case '=':
				case '-':
					if (gnMode==MENU_TRACK_MODE) 
						break;

					if (wParam=='-') gnZoomIndex--;
					else gnZoomIndex++;

		            gnZoomIndex = BOUND(gnZoomIndex, MIN_ZOOM_INDEX, MAX_ZOOM_INDEX);
					gnZoom = gnZoomFactors[gnZoomIndex];
					ResetTitle();
					CalcZoomedSize();

					break;
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					if (gnMode==MENU_TRACK_MODE) 
						break;
					gnZoom = wParam - '0';
					ResetTitle();
					break;
				case '0':
					if (gnMode==MENU_TRACK_MODE) 
						break;
					gnZoom = 10;
					ResetTitle();
					break;

				case 'a':
					SendMessage(hwnd, WM_COMMAND, MENU_ARROW_MODE, 0L);
					break;
				case 'b':
					SendMessage(hwnd, WM_COMMAND, MENU_BOUND, 0L);
					break;
				case 'c':
					SendMessage(hwnd, WM_COMMAND, MENU_CENTER_HIGHLIGHT_ENABLED, 0L);
					break;
				case 'f':
					SaveRegistry(hwnd);
					break;
				case 'H':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_HOR_STEP, 0L);
					break;
				case 'h':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_HOR_LINEAR, 0L);
					break;
				case 'i':
					SendMessage(hwnd, WM_COMMAND, MENU_STATISTICS_INFO, 0L);
					break;
				case 'l':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_HIGHLIGHT, 0L);
					break;
				case 'm':
					SendMessage(hwnd, WM_COMMAND, MENU_MOUSE_MODE, 0L);
					break;
				case 'o':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_OFF, 0L);
					break;
				case 'q':
					SendMessage(hwnd, WM_COMMAND, MENU_EDIT_QUIT, 0L);
					break;
				case 'r':
					SendMessage(hwnd, WM_COMMAND, MENU_RULER_ENABLED, 0L);
					break;
				case 's':
					SendMessage(hwnd, WM_COMMAND, MENU_SCREEN, 0L);
					break;
				case 'u':
					SendMessage(hwnd, WM_COMMAND, MENU_EDIT_REFRESH, 0L);
					break;
				case 'V':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_VER_STEP, 0L);
					break;
				case 'v':
					SendMessage(hwnd, WM_COMMAND, MENU_GRAPH_VER_LINEAR, 0L);
					break;
				case 'w':
					SendMessage(hwnd, WM_COMMAND, MENU_ACTIVE_WINDOW, 0L);
					break;
				case 'x':
					SendMessage(hwnd, WM_COMMAND, MENU_DISPLAY_HEX, 0L);
					break;
				case 'y':
					SendMessage(hwnd, WM_COMMAND, MENU_INVERSE_Y, 0L);
					break;
				case '#':
					SendMessage(hwnd, WM_COMMAND, MENU_GRID_ENABLED, 0L);
					break;
			}

			if (wParam>='0' && wParam<='9') {
				gnZoomIndex = findIndex(gnZoom);
				CalcZoomedSize();
			}

            DoTheZoomIn(NULL);

			break;

        case WM_COMMAND:
            switch (LOWORD(wParam)) {
				case MENU_FILTER_SNOOP:
					CheckMenuItem(hmenu, gnDisplayMode, MF_UNCHECKED);
					gnDisplayMode=MENU_FILTER_SNOOP;
					CheckMenuItem(hmenu, gnDisplayMode, MF_CHECKED);
					DoTheZoomIn(NULL);
					break;

				case MENU_FILTER_GRADIENT:
					if (gnColorBits<24) break;
					CheckMenuItem(hmenu, gnDisplayMode, MF_UNCHECKED);
					gnDisplayMode=MENU_FILTER_GRADIENT;
					CheckMenuItem(hmenu, gnDisplayMode, MF_CHECKED);
					DoTheZoomIn(NULL);
				break;

                case MENU_EDIT_COPY:
                    CopyToClipboard();	
                    break;

                case MENU_EDIT_REFRESH:	
//					if (gnMode==MENU_MOUSE_MODE)
						GetCursorPos(&gptZoom);

                    DoTheZoomIn(NULL);
                    break;

				case MENU_EDIT_SAVE_CONFIG:
					SaveRegistry(hwnd);
					break;

				case MENU_EDIT_AUTO_SAVE_CONFIG:
					CheckMenuItem(hmenu, MENU_EDIT_AUTO_SAVE_CONFIG, (gbSaveConfigExit ? MF_UNCHECKED : MF_CHECKED));
                    gbSaveConfigExit = !gbSaveConfigExit;
					break;

                case MENU_EDIT_QUIT:
					if (ghpalPhysical)
						DeleteObject(ghpalPhysical);

					if (gbSaveConfigExit)
						SaveRegistry(hwnd);

					SaveAutoSaveRegistry();

					DestroyWindow(hwnd);

                    break;

				case MENU_GRAPH_OFF:
				case MENU_GRAPH_HOR_STEP:
				case MENU_GRAPH_HOR_LINEAR:
				case MENU_GRAPH_VER_STEP:
				case MENU_GRAPH_VER_LINEAR:
				case MENU_GRAPH_STATISTICS:
					if (LOWORD(wParam)==gnRGBGraphMode && gnRGBGraphMode!=MENU_GRAPH_OFF) {
						CheckMenuItem(hmenu, gnRGBGraphMode, MF_UNCHECKED);
						gnRGBGraphMode = MENU_GRAPH_OFF;
						CheckMenuItem(hmenu, gnRGBGraphMode, MF_CHECKED);
					} else {
						CheckMenuItem(hmenu, gnRGBGraphMode, MF_UNCHECKED);
						gnRGBGraphMode = LOWORD(wParam);
						CheckMenuItem(hmenu, gnRGBGraphMode, MF_CHECKED);
					}

					switch (gnRGBGraphMode) {
						case MENU_GRAPH_OFF:
							EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_GRAYED);
							EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_GRAYED);
							break;

						case MENU_GRAPH_HOR_STEP:
						case MENU_GRAPH_HOR_LINEAR:
						case MENU_GRAPH_VER_STEP:
						case MENU_GRAPH_VER_LINEAR:
							EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_ENABLED);
							EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_GRAYED);
							break;

						case MENU_GRAPH_STATISTICS:
							EnableMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, MF_GRAYED);
							EnableMenuItem(hmenu, MENU_STATISTICS_INFO, MF_ENABLED);
							break;
					}

					break;

				case MENU_GRAPH_HIGHLIGHT:
					if (gnRGBGraphMode!=MENU_GRAPH_OFF && gnRGBGraphMode!=MENU_GRAPH_STATISTICS) {
						gbRGBGraphHighlight = !gbRGBGraphHighlight;
						CheckMenuItem(hmenu, MENU_GRAPH_HIGHLIGHT, (gbRGBGraphHighlight ? MF_CHECKED : MF_UNCHECKED));
					}

					break;

                case MENU_OPTIONS_REFRESHRATE:
                    DialogBox(ghInst, (LPSTR)MAKEINTRESOURCE(DID_REFRESHRATE),
                            hwnd, (DLGPROC)RefreshRateDlgProc);

                    break;

                case MENU_OPTIONS_HISTOGRAM:
                    DialogBox(ghInst, (LPSTR)MAKEINTRESOURCE(DID_HISTOGRAM),
                            hwnd, (DLGPROC)HistogramDlgProc);

                    break;

				case MENU_OPTIONS_ONTOP:
					SetWindowPos(hwnd, (gbOnTop ? HWND_NOTOPMOST : HWND_TOPMOST), 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
					CheckMenuItem(hmenu, MENU_OPTIONS_ONTOP, (gbOnTop ? MF_UNCHECKED : MF_CHECKED));
					gbOnTop = !gbOnTop;
					break;

                case MENU_GRID_ENABLED:
					CheckMenuItem(hmenu, MENU_GRID_ENABLED, (gbGrid ? MF_UNCHECKED : MF_CHECKED));
                    gbGrid = !gbGrid;
                    DoTheZoomIn(NULL);
                    break;

				case MENU_RULER_ENABLED:
					CheckMenuItem(hmenu, MENU_RULER_ENABLED, (gbRuler ? MF_UNCHECKED : MF_CHECKED));
                    gbRuler = !gbRuler;

					// recalculate the size
					CalcZoomedSize();
					CalcRGBGraphSize();

                    DoTheZoomIn(NULL);
					break;

                case MENU_CENTER_HIGHLIGHT_ENABLED:
					CheckMenuItem(hmenu, MENU_CENTER_HIGHLIGHT_ENABLED, (gbCenterHighlight ? MF_UNCHECKED : MF_CHECKED));
                    gbCenterHighlight = !gbCenterHighlight;
                    DoTheZoomIn(NULL);
                    break;

				case MENU_STATISTICS_INFO:
					if (gnRGBGraphMode==MENU_GRAPH_STATISTICS) {
						CheckMenuItem(hmenu, MENU_STATISTICS_INFO, (gbStatisticsInfo ? MF_UNCHECKED : MF_CHECKED));
	                    gbStatisticsInfo = !gbStatisticsInfo;
		                DoTheZoomIn(NULL);
					}
					break;

				case MENU_SCREEN:
				case MENU_ACTIVE_WINDOW:
					if (gnCoordSystem==MENU_SCREEN) {
						CheckMenuItem(hmenu, MENU_SCREEN, MF_UNCHECKED);
						CheckMenuItem(hmenu, MENU_ACTIVE_WINDOW, MF_CHECKED);
					} else {
						CheckMenuItem(hmenu, MENU_SCREEN, MF_CHECKED);
						CheckMenuItem(hmenu, MENU_ACTIVE_WINDOW, MF_UNCHECKED);
					}
					gnCoordSystem=LOWORD(wParam);
					break;

				case MENU_INVERSE_Y:
					gbInverseY=!gbInverseY;
					CheckMenuItem(hmenu, MENU_INVERSE_Y, gbInverseY?MF_CHECKED:MF_UNCHECKED);
					break;

				case MENU_DISPLAY_HEX:
					gbDisplayHex=!gbDisplayHex;
					CheckMenuItem(hmenu, MENU_DISPLAY_HEX, gbDisplayHex?MF_CHECKED:MF_UNCHECKED);
					break;

				case MENU_BOUND:
					gbBound=!gbBound;
					CheckMenuItem(hmenu, MENU_BOUND, gbBound?MF_CHECKED:MF_UNCHECKED);
					break;

				case MENU_MOUSE_MODE:
				case MENU_ARROW_MODE:
					
					gnMode = LOWORD(wParam);

					CheckMenuItem(hmenu, MENU_MOUSE_MODE, (gnMode==MENU_MOUSE_MODE ? MF_CHECKED : MF_UNCHECKED));
					CheckMenuItem(hmenu, MENU_ARROW_MODE, (gnMode==MENU_ARROW_MODE ? MF_CHECKED : MF_UNCHECKED));

					ResetTitle();
					break;

                case MENU_HELP_ABOUT:
                    DialogBox(ghInst, (LPSTR)MAKEINTRESOURCE(DID_ABOUT),
                            hwnd, (DLGPROC)AboutDlgProc);

                    break;

                default:
                     break;
            }

            break;

        case WM_CLOSE:
            if (ghpalPhysical)
                DeleteObject(ghpalPhysical);

            DestroyWindow(hwnd);

            break;

        case WM_DESTROY:
			// make sure that timer event won't happen afterward
			gbQuit = TRUE;

			if (gfRefEnable)
				KillTimer(ghwndApp, IDTIMER_SNOOP);

			DeleteObject(ghRPen);
			DeleteObject(ghGPen);
			DeleteObject(ghBPen);

			DeleteObject(ghR2Pen);

			DeleteObject(ghGrayPen);
			DeleteObject(ghGrayBrush);

			DeleteObject(ghPeakFont);

			if (ghrgnCanvasClip!=0) DeleteObject(ghrgnCanvasClip);
			if (ghrgnHorGraphClip!=0) DeleteObject(ghrgnHorGraphClip);
			if (ghrgnVerGraphClip!=0) DeleteObject(ghrgnVerGraphClip);
			if (ghrgnHistogram!=0) DeleteObject(ghrgnHistogram);

			free(gcolorBuf);
			free(gptBuf[0]);
			free(gptBuf[1]);
			free(gptBuf[2]);

            PostQuitMessage(0);
            break;

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }

    return 0L;
}

/************************************************************************
* CalcZoomedSize
*
* Calculates some globals.  This routine needs to be called any
* time that the size of the app or the zoom factor changes.
*
* History:
*
************************************************************************/

VOID CalcZoomedSize(VOID)
{
    RECT rc;

    GetClientRect(ghwndApp, &rc);

	if (gbRuler) {
		gcxZoomed = ((rc.right-RULER_HEIGHT)/gnZoom) + 1;
		gcyZoomed = ((rc.bottom-SB_HEIGHT-RULER_HEIGHT)/gnZoom) + 1;
	} else {
		gcxZoomed = (rc.right/gnZoom) + 1;
		gcyZoomed = ((rc.bottom-SB_HEIGHT)/gnZoom) + 1;
	}
}

/************************************************************************
* CalcRGBGraphSize
*
* Calculates some globals.  This routine needs to be called any
* time that the size of the app.
*
* History: 09/06/2000 Created.
*
************************************************************************/

VOID CalcRGBGraphSize(VOID)
{
	RECT rc;
	int  y=0;

    GetClientRect(ghwndApp, &rc);

	if (gbRuler) {
		rc.right = rc.right - RULER_HEIGHT;
		rc.bottom = rc.bottom - RULER_HEIGHT;
	}

	gnRGBGraphWidth = (int)(rc.right*0.33);
	gfRGBVerFactor = (gnRGBGraphWidth-INSET_BORDER-4)/255.0f;
	if (gfRGBVerFactor<0) gfRGBVerFactor=0;

	gnRGBGraphHeight = (int)((rc.bottom-SB_HEIGHT)*0.33);
	gfRGBHorFactor = (gnRGBGraphHeight-INSET_BORDER-4)/255.0f;
	if (gfRGBHorFactor<0) gfRGBHorFactor=0;

	gnHorGraphBase = gnRGBGraphHeight;
	gnHorGraphTop = INSET_BORDER;

	gnVerGraphBase = rc.right - gnRGBGraphWidth;
	gnVerGraphTop = rc.right - INSET_BORDER;

	grectHorGraph.left	= rc.left + INSET_BORDER;
	grectHorGraph.right = rc.right - INSET_BORDER;
	grectHorGraph.top	= INSET_BORDER;
	grectHorGraph.bottom= gnRGBGraphHeight;

	grectVerGraph.left	= rc.right - gnRGBGraphWidth;
	grectVerGraph.right = rc.right - INSET_BORDER;
	grectVerGraph.top	= rc.top + INSET_BORDER;
	grectVerGraph.bottom= rc.bottom - SB_HEIGHT - INSET_BORDER;

	if (ghrgnCanvasClip!=0) DeleteObject(ghrgnCanvasClip);

	// clip coordinate is physical coord, SetWindowOrgExt don't have effect on it!
	if (gbRuler) 
		ghrgnCanvasClip=CreateRectRgn(rc.left+RULER_HEIGHT, rc.top+RULER_HEIGHT, 
									  rc.right+RULER_HEIGHT, rc.bottom-SB_HEIGHT+RULER_HEIGHT);
	else
		ghrgnCanvasClip=CreateRectRgn(rc.left, rc.top, rc.right, rc.bottom - SB_HEIGHT);

	if (ghrgnHorGraphClip!=0) DeleteObject(ghrgnHorGraphClip);
	if (gbRuler) 
		ghrgnHorGraphClip=CreateRectRgn(grectHorGraph.left+RULER_HEIGHT, grectHorGraph.top+RULER_HEIGHT,
	   								    grectHorGraph.right+RULER_HEIGHT, grectHorGraph.bottom+RULER_HEIGHT);	
	else 
		ghrgnHorGraphClip=CreateRectRgn(grectHorGraph.left, grectHorGraph.top,
	   								    grectHorGraph.right, grectHorGraph.bottom);	

	if (ghrgnVerGraphClip!=0) DeleteObject(ghrgnVerGraphClip);
	if (gbRuler) 
		ghrgnVerGraphClip=CreateRectRgn(grectVerGraph.left+RULER_HEIGHT, grectVerGraph.top+RULER_HEIGHT,
									    grectVerGraph.right+RULER_HEIGHT, grectVerGraph.bottom+RULER_HEIGHT);
	else
		ghrgnVerGraphClip=CreateRectRgn(grectVerGraph.left, grectVerGraph.top,
									    grectVerGraph.right, grectVerGraph.bottom);

	// calc histogram box
	// calc delta
	gnHistogramWidth=rc.right-2*INSET_BORDER-HISTOGRAM_RULER_HEIGHT;

	if (gnHistogramWidth<128) {
		gnHistogramWidth=64;
		gnHistogramDelta=4;
		gnHistogramGridX=5;		// 5*4=20
		gnHistogramHeight=gnHistogramWidth/2;
		if (gnHistogramHeight>gnRGBGraphHeight)
			gnHistogramHeight=gnRGBGraphHeight;
		if (gnHistogramHeight>32+HISTOGRAM_RULER_HEIGHT) 
			gnHistogramHeight=32+HISTOGRAM_RULER_HEIGHT;
		if (gnHistogramHeight<HISTOGRAM_RULER_HEIGHT+16)
			gnHistogramHeight=HISTOGRAM_RULER_HEIGHT+16;
	} else if (gnHistogramWidth<256) {
		gnHistogramWidth=128;
		gnHistogramDelta=2;
		gnHistogramGridX=5;		// 5*2=10
		gnHistogramHeight=gnHistogramWidth/2;
		if (gnHistogramHeight>gnRGBGraphHeight)
			gnHistogramHeight=gnRGBGraphHeight;
		if (gnHistogramHeight>64+HISTOGRAM_RULER_HEIGHT) 
			gnHistogramHeight=64+HISTOGRAM_RULER_HEIGHT;
		if (gnHistogramHeight<HISTOGRAM_RULER_HEIGHT+16)
			gnHistogramHeight=HISTOGRAM_RULER_HEIGHT+16;
	} else {
		gnHistogramWidth=256;
		gnHistogramDelta=1;
		gnHistogramGridX=10;	// 10*1=10
		gnHistogramHeight=gnHistogramWidth/2;
		if (gnHistogramHeight>gnRGBGraphHeight)
			gnHistogramHeight=gnRGBGraphHeight;
		if (gnHistogramHeight>128+HISTOGRAM_RULER_HEIGHT) 
			gnHistogramHeight=128+HISTOGRAM_RULER_HEIGHT;
		if (gnHistogramHeight<HISTOGRAM_RULER_HEIGHT+16)
			gnHistogramHeight=HISTOGRAM_RULER_HEIGHT+16;
	}

	// clac height
	gnHistogramOrgX=rc.right/2-(HISTOGRAM_RULER_HEIGHT + 256/gnHistogramDelta+2)/2 + HISTOGRAM_RULER_HEIGHT;
	gnHistogramOrgY=INSET_BORDER;
	
	grectHistogram.left	  = gnHistogramOrgX-HISTOGRAM_RULER_HEIGHT;
	grectHistogram.right  = grectHistogram.left + HISTOGRAM_RULER_HEIGHT + 256/gnHistogramDelta+2;
	grectHistogram.top	  = INSET_BORDER;
	grectHistogram.bottom = grectHistogram.top + HISTOGRAM_RULER_HEIGHT + gnHistogramHeight;	

	if (ghrgnHistogram!=0) DeleteObject(ghrgnHistogram);
	if (gbRuler) 
		ghrgnHistogram=CreateRectRgn(grectHistogram.left+RULER_HEIGHT, grectHistogram.top+RULER_HEIGHT,
									    grectHistogram.right+RULER_HEIGHT, grectHistogram.bottom+RULER_HEIGHT);
	else
		ghrgnHistogram=CreateRectRgn(grectHistogram.left, grectHistogram.top,
									    grectHistogram.right, grectHistogram.bottom);	

	if (gnHistogramDelta==1) {
		y=rc.bottom-SB_HEIGHT-TEXTBOX_HEIGHT_2_ROWS-INSET_BORDER;

		gnHistogramTextX=rc.right/2-TEXTBOX_WIDTH_2_COLS/2;
	} else if (gnHistogramDelta==2) {
		y=rc.bottom-SB_HEIGHT-TEXTBOX_HEIGHT_4_ROWS-INSET_BORDER;

		gnHistogramTextX=rc.right/2-TEXTBOX_WIDTH_1_COLS/2;
	}
	if (y<(gcyZoomed/2+1)*gnZoom)
		gnHistogramTextY = (gcyZoomed/2+1)*gnZoom;
	else
		gnHistogramTextY = y;

/*
grectHistogram.left	  = 0;
grectHistogram.right  = 256 ;
grectHistogram.top	  = 0;
grectHistogram.bottom = 64;	
*/
}

/************************************************************************
* DoTheGrid  <Specific to the ZoominEx>
*
* Draw the grid. If the zoom is less than 2, the grid is not drawing.
*
* History: 01/06/99 Created
*
************************************************************************/

void DoTheGrid(HDC hdc, HGDIOBJ hPen)
{
    RECT rc;
    int iX;
    int iY;

    if (gnZoom < 2)
        return;

    GetClientRect(ghwndApp, &rc);

	SelectObject(hdc, hPen);

    for (iX = 0; iX < gcxZoomed; iX++) {
        MoveToEx(hdc, iX * gnZoom, rc.top, NULL);
        LineTo(hdc, iX * gnZoom, rc.bottom);
    }

    for (iY = 0; iY < gcyZoomed; iY++) {
        MoveToEx(hdc, rc.left, iY * gnZoom, NULL);
        LineTo(hdc, rc.right, iY * gnZoom);
    }
}

void DoTheRuler(HDC hdc)
{
    RECT rc;
    int iX;
    int iY;

    GetClientRect(ghwndApp, &rc);

	SelectObject(hdc, ghGrayPen);
	SelectObject(hdc, ghGrayBrush);

	Rectangle(hdc, rc.left, rc.top, RULER_HEIGHT, rc.bottom-SB_HEIGHT);
	Rectangle(hdc, rc.left+RULER_HEIGHT, rc.top, rc.right, RULER_HEIGHT);

	SelectObject(hdc, GetStockObject(BLACK_PEN));

    MoveToEx(hdc, RULER_HEIGHT, RULER_HEIGHT-1, NULL);
    LineTo(hdc, rc.right, RULER_HEIGHT-1);
    for (iX = 0; iX < gcxZoomed; iX+=5) {
        MoveToEx(hdc, iX * gnZoom + RULER_HEIGHT, RULER_HEIGHT-1, NULL);
        LineTo(hdc, iX * gnZoom + RULER_HEIGHT, ((iX%10)?5:1));
    }

	MoveToEx(hdc, RULER_HEIGHT-1, RULER_HEIGHT, NULL);
    LineTo(hdc, RULER_HEIGHT-1, rc.bottom);
    for (iY = 0; iY < gcyZoomed; iY+=5) {
        MoveToEx(hdc, RULER_HEIGHT-1, iY * gnZoom + RULER_HEIGHT, NULL);
        LineTo(hdc, ((iY%10)?5:1), iY * gnZoom + RULER_HEIGHT);
    }

}

void DoTheCenterHightlight(HDC hdc)
{
	POINT pt[5];
	HPEN hPen;

    if (gnZoom < 2)
        return;

	pt[0].x=(gcxZoomed/2)*gnZoom;
	pt[0].y=(gcyZoomed/2)*gnZoom;

	pt[1].x=(gcxZoomed/2+1)*gnZoom;
	pt[1].y=(gcyZoomed/2)*gnZoom;

	pt[2].x=(gcxZoomed/2+1)*gnZoom;
	pt[2].y=(gcyZoomed/2+1)*gnZoom;

	pt[3].x=(gcxZoomed/2)*gnZoom;
	pt[3].y=(gcyZoomed/2+1)*gnZoom;

	pt[4].x=(gcxZoomed/2)*gnZoom;
	pt[4].y=(gcyZoomed/2)*gnZoom;

	hPen = (HPEN)SelectObject(hdc, ghRPen);
	Polyline(hdc, pt, 5);
	SelectObject(hdc, hPen);
}

/************************************************************************
* DoTheRGBGraph
*
* Does the painting of RGB graph
*
* Arguments:
*   HDC hdc - DCto draw on.
*   HDC hdcScreen - screen DC          
*
* History:
*
************************************************************************/
// TODO : unnecessary function!
void selectPen(HDC hdc, int c)
{
	switch (c) {
		case 0:
			SelectObject( hdc, ghRPen);
			break;
		case 1:
			SelectObject( hdc, ghGPen);
			break;
		case 2:
			SelectObject( hdc, ghBPen);
			break;
	}
}

VOID FillinColorBuffer(HDC hdcScreen)
{
	int x, y, i;
/*
	unsigned char *pData;

	if (gnColorBits>=24) {
		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
			pData=gpCaptureBuf+gnCaptureMemoryWidth*(gnCaptureHeight-1-gcyZoomed/2);
			for (i=0; i<gcxZoomed; i++) {
				gcolorBuf[i]=(pData[0]<<16)+(pData[1]<<8)+pData[2];	// 00bbggrr
				if (gnColorBits==32) pData+=4;
				else pData+=3;
			}
		} else {
			pData=gpCaptureBuf+gnCaptureMemoryWidth*(gnCaptureHeight-1)+(gnColorBits/8)*(gcxZoomed/2);
			for (i=0; i<gcyZoomed; i++) {
				gcolorBuf[i]=(pData[0]<<16)+(pData[1]<<8)+pData[2];	// 00bbggrr
				pData-=gnCaptureMemoryWidth;
			}
		}
	} else {*/
 
 	// hdcScreen is assumed to be the zoomed and streched DC
 	if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
 		for (i=0; i<gcxZoomed; i++)
 			gcolorBuf[i]=GetPixel(hdcScreen, i*gnZoom, gcyZoomed*gnZoom/2);
 	} else {
 		for (i=0; i<gcyZoomed; i++)
 			gcolorBuf[i]=GetPixel(hdcScreen, gcxZoomed*gnZoom/2, i*gnZoom);
 	}
 	return;
 
/*	
// always use windows API to avoid the problem on Brian's laptop
		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
			x = gnx - gcxZoomed / 2;
			for (i=0; i<gcxZoomed; i++) {
				gcolorBuf[i]=GetPixel(hdcScreen, x, gny);
				x++;
			}
		} else {
			y = gny - gcyZoomed / 2;
			for (i=0; i<gcyZoomed; i++) {
				gcolorBuf[i]=GetPixel(hdcScreen, gnx, y);
				y++;
			}
		}
//	}
*/
}

#define GITTER		40

VOID DoTheHighlight(HDC hdc, HDC hdcScreen)
{
//	HPALETTE hpalMemHilightOld = NULL;
	HDC hdcMemHilight;
	HBITMAP hBmpMemHilight;
	HBITMAP hBmpMemHilightOld;
	int x, y, i;
//	COLORREF rgb;
	int r, g, b;

	hdcMemHilight = CreateCompatibleDC(hdc);
	if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR)
		hBmpMemHilight = CreateCompatibleBitmap(hdc, gcxZoomed, 1);
	else
		hBmpMemHilight = CreateCompatibleBitmap(hdc, 1, gcyZoomed);
	hBmpMemHilightOld = (HBITMAP)SelectObject(hdcMemHilight, hBmpMemHilight);

	// pallete?

	if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
		for (x=-gcxZoomed/2, i=0; x<gcxZoomed/2; x++, i++) {
			if (x!=0) {
				r=GetRValue(gcolorBuf[i]);
				g=GetGValue(gcolorBuf[i]);		
				b=GetBValue(gcolorBuf[i]);

/*				if (r==255 && g==255 && b==255) {
					r=g=b=255-20;
				} else {*/
					if (r<255-GITTER) r+=GITTER;
					else r=(int)(r*((255-GITTER)/255.0))+GITTER;
					if (g<255-GITTER) g+=GITTER;
					else g=(int)(g*((255-GITTER)/255.0))+GITTER;
					if (b<255-GITTER) b+=GITTER;
					else b=(int)(b*((255-GITTER)/255.0))+GITTER;
//				}

				SetPixel(hdcMemHilight, x+gcxZoomed / 2, 0, RGB(r, g, b));
			} else 
				SetPixel(hdcMemHilight, x+gcxZoomed / 2, 0, gcolorBuf[i]);
		}
	} else {
		for (y=-gcyZoomed/2, i=0; y<gcyZoomed/2; y++, i++) {
			if (y!=0) {
				r=GetRValue(gcolorBuf[i]);
				g=GetGValue(gcolorBuf[i]);		
				b=GetBValue(gcolorBuf[i]);

/*				if (r==255 && g==255 && b==255) {
					r=g=b=255-20;
				} else {*/
					if (r<255-GITTER) r+=GITTER;
					else r=(int)(r*((255-GITTER)/255.0))+GITTER;
					if (g<255-GITTER) g+=GITTER;
					else g=(int)(g*((255-GITTER)/255.0))+GITTER;
					if (b<255-GITTER) b+=GITTER;					
					else b=(int)(b*((255-GITTER)/255.0))+GITTER;
//				}

				SetPixel(hdcMemHilight, 0, y+gcyZoomed/2, RGB(r, g, b));
			} else 
				SetPixel(hdcMemHilight, 0, y+gcyZoomed/2, gcolorBuf[i]);
		}
	}

	if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR)
		StretchBlt(hdc, 0, (gcyZoomed / 2)*gnZoom, gnZoom * gcxZoomed, gnZoom,
			       hdcMemHilight, 0, 0, gcxZoomed, 1, SRCCOPY);
	else 
		StretchBlt(hdc, (gcxZoomed / 2)*gnZoom, 0, gnZoom, gnZoom * gcyZoomed, 
			       hdcMemHilight, 0, 0, 1, gcyZoomed, SRCCOPY);

	SelectObject(hdcMemHilight, hBmpMemHilightOld);
	DeleteObject(hBmpMemHilight);
	DeleteDC(hdcMemHilight);	
}

VOID DoTheRGBGraph(HDC hdc, HDC hdcScreen)
{
	int i;
	int x, y;
	int c;

	switch (gnRGBGraphMode) {
	case MENU_GRAPH_HOR_STEP:
	case MENU_GRAPH_HOR_LINEAR:
		// select clip region
		SelectClipRgn(hdc, ghrgnHorGraphClip);

		FillRect( hdc, &grectHorGraph, (HBRUSH)GetStockObject(BLACK_BRUSH));

		if (gbRGBGraphHighlight && gnColorBits>=24) {
			SelectObject(hdc, GetStockObject(NULL_PEN));
			SelectObject(hdc, GetStockObject(DKGRAY_BRUSH));
			Rectangle(hdc, gnZoom*(gcxZoomed/2)-1, gnHorGraphTop+1,
						   gnZoom*(gcxZoomed/2+1)+1, gnHorGraphBase-1);
		}

		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP) {
			//prepare points
			x = 0;
			for (i=0; i<gcxZoomed; i++) {
				gptBuf[0][i].x = gptBuf[1][i].x = gptBuf[2][i].x = x;
				gptBuf[0][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetRValue(gcolorBuf[i])) - 2;
				gptBuf[1][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetGValue(gcolorBuf[i])) - 2;
				gptBuf[2][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetBValue(gcolorBuf[i])) - 2;
				x += gnZoom;
			}
			// color blending
			SetROP2(hdc, R2_MERGEPEN);
			// draw it
			for (c=0; c<3; c++) {
				selectPen(hdc, c);
				for (i=0; i<gcxZoomed; i++) {
					MoveToEx(hdc, gptBuf[c][i].x,        gptBuf[c][i].y, NULL);
					LineTo  (hdc, gptBuf[c][i].x+gnZoom, gptBuf[c][i].y);
				}
			}
			SetROP2(hdc, R2_COPYPEN);
		} else {
			//prepare points
			x = gnZoom/2;
			for (i=0; i<gcxZoomed; i++) {
				gptBuf[0][i].x = gptBuf[1][i].x = gptBuf[2][i].x = x;
				gptBuf[0][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetRValue(gcolorBuf[i])) - 2;
				gptBuf[1][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetGValue(gcolorBuf[i])) - 2;
				gptBuf[2][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetBValue(gcolorBuf[i])) - 2;
				x += gnZoom;
			}
			// color blending
			SetROP2(hdc, R2_MERGEPEN);
			// draw it
			for (c=0; c<3; c++) {
				selectPen(hdc, c);
				Polyline(hdc, gptBuf[c], gcxZoomed);
			}
			SetROP2(hdc, R2_COPYPEN);
		}

		// draw outside
		SelectObject(hdc, GetStockObject(WHITE_PEN));
		SelectObject(hdc, GetStockObject(NULL_BRUSH));
		Rectangle(hdc, grectHorGraph.left, grectHorGraph.top, grectHorGraph.right, grectHorGraph.bottom);
		break;

	case MENU_GRAPH_VER_STEP:
	case MENU_GRAPH_VER_LINEAR:
		// select clip region
		SelectClipRgn(hdc, ghrgnVerGraphClip);

		FillRect( hdc, &grectVerGraph, (HBRUSH)GetStockObject(BLACK_BRUSH));

		if (gbRGBGraphHighlight && gnColorBits>=24) {
			SelectObject(hdc, GetStockObject(NULL_PEN));
			SelectObject(hdc, GetStockObject(DKGRAY_BRUSH));
			Rectangle(hdc, gnVerGraphBase+1, gnZoom*(gcyZoomed/2)-1, 
						   gnVerGraphTop-1, gnZoom*(gcyZoomed/2+1)+1);
		}

		if (gnRGBGraphMode==MENU_GRAPH_VER_STEP) {
			//prepare points
			y = 0;
			for (i=0; i<gcyZoomed; i++) {
				gptBuf[0][i].y = gptBuf[1][i].y = gptBuf[2][i].y = y;
				gptBuf[0][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetRValue(gcolorBuf[i])) + 2;
				gptBuf[1][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetGValue(gcolorBuf[i])) + 2;
				gptBuf[2][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetBValue(gcolorBuf[i])) + 2;
				y += gnZoom;
			}
			// color blending
			SetROP2(hdc, R2_MERGEPEN);
			// draw it
			for (c=0; c<3; c++) {
				selectPen(hdc, c);
				for (i=0; i<gcyZoomed; i++) {
					MoveToEx(hdc, gptBuf[c][i].x, gptBuf[c][i].y, NULL);
					LineTo  (hdc, gptBuf[c][i].x, gptBuf[c][i].y+gnZoom);
				}
			}
			SetROP2(hdc, R2_COPYPEN);
		} else {
			//prepare points
			y = gnZoom/2;
			for (i=0; i<gcyZoomed; i++) {
				gptBuf[0][i].y = gptBuf[1][i].y = gptBuf[2][i].y = y;
				gptBuf[0][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetRValue(gcolorBuf[i])) + 2;
				gptBuf[1][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetGValue(gcolorBuf[i])) + 2;
				gptBuf[2][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetBValue(gcolorBuf[i])) + 2;
				y += gnZoom;
			}
			// color blending
			SetROP2(hdc, R2_MERGEPEN);
			// draw it
			for (c=0; c<3; c++) {
				selectPen(hdc, c);
				Polyline(hdc, gptBuf[c], gcyZoomed);
			}
			SetROP2(hdc, R2_COPYPEN);
		}

		// draw outside
		SelectObject(hdc, GetStockObject(WHITE_PEN));
		SelectObject(hdc, GetStockObject(NULL_BRUSH));
		Rectangle(hdc, grectVerGraph.left, grectVerGraph.top, grectVerGraph.right, grectVerGraph.bottom);

		break;
	case MENU_GRAPH_STATISTICS:
		
		break;
	}

	SelectClipRgn(hdc, ghrgnCanvasClip);
}

int count[3][256];

VOID DoTheStatistics(HDC hdc, HDC hdcScreen)
{
	int x, y;
//	COLORREF rgb;
	int r, g, b;
	int max[3], min[3];
	int med[3], avg[3];
	int maxc[3];
	int maxc_color[3];
	int realmaxc;
	int c, i;
	int cy, cx;
	float fy;
	char szBuffer[128];
	int total;
	int acc;
	float sum;
	POINT ptHist[256];

	int nMemoryWidth;
	int width, height;
	int padding;

	unsigned char *pData;
	unsigned char *buf;

	HFONT hOldFont;

	if (gnColorBits < 24) return;

	width=gcxZoomed;
	height=gcyZoomed;
	// we always capture screen for true-color display, don't have to captur again, just use it
/*
	ScreenCapture(hdcScreen, gnx-gcxZoomed/2, gny-gcyZoomed/2, width, height,
				  &buf, &padding);
*/
	buf = gpCaptureBuf;
	padding = gnCapturePadding;
	nMemoryWidth = gnCaptureMemoryWidth;
	
	// init
	memset(count, 0, 3*256*sizeof(int));
	max[0]=max[1]=max[2]=0;
	min[0]=min[1]=min[2]=255;
	maxc[0]=maxc[1]=maxc[2]=0;

	pData=buf;

	for (y=gny-gcyZoomed/2, cy=0; cy<gcyZoomed; y++, cy++) {
		for (x=gnx-gcxZoomed/2, cx=0; cx<gcxZoomed; x++, cx++) {

// this is too slow 
/*
//			rgb=GetPixel(hdcScreen, x, (gny+gcyZoomed/2-1)-y);
			rgb=GetPixel(hdcScreen, x, y);
			r=GetRValue(rgb);
			g=GetGValue(rgb);
			b=GetBValue(rgb);
*/
			b=*(pData++);
			g=*(pData++);
			r=*(pData++);

			if (gnColorBits==32) pData++;

			count[0][r]++;
			count[1][g]++;
			count[2][b]++;

			if (gbStatisticsInfo) {
				if (r<min[0]) min[0]=r;
				if (r>max[0]) max[0]=r;
				if (g<min[1]) min[1]=g;
				if (g>max[1]) max[1]=g;
				if (b<min[2]) min[2]=b;
				if (b>max[2]) max[2]=b;
			}

			if (count[0][r]>=maxc[0]) {
				maxc[0]=count[0][r];
				maxc_color[0]=r;
			}
			if (count[1][g]>=maxc[1]) {
				maxc[1]=count[1][g];
				maxc_color[1]=g;
			}
			if (count[2][b]>=maxc[2]) {
				maxc[2]=count[2][b];
				maxc_color[2]=b;
			}
		}
		pData+=padding;
	}

	// draw the diagram
	SelectClipRgn(hdc, ghrgnHistogram);

	FillRect( hdc, &grectHistogram, (HBRUSH)GetStockObject(BLACK_BRUSH));

	// draw the ruler
	SelectObject(hdc, GetStockObject(WHITE_PEN));
	// horizontal grid
	MoveToEx( hdc, gnHistogramOrgX-HISTOGRAM_RULER_HEIGHT, gnHistogramOrgY+gnHistogramHeight, NULL );
	LineTo  ( hdc, gnHistogramOrgX+gnHistogramWidth, gnHistogramOrgY+gnHistogramHeight );
	for (x=0; x<gnHistogramWidth; x+=gnHistogramGridX) {
		MoveToEx( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight, NULL );
		if ( (gnHistogramDelta==1 && (x%50)==0) || (gnHistogramDelta!=1 && x%25==0) ) 
			LineTo( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight+6 );
		else
			LineTo( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight+4 );
	}
	// vertical grid
	MoveToEx( hdc, gnHistogramOrgX, gnHistogramOrgY, NULL );
	LineTo  ( hdc, gnHistogramOrgX, gnHistogramOrgY+gnHistogramHeight+HISTOGRAM_RULER_HEIGHT );
	for (fy=0; fy<=1.0; fy+=0.1f) {
		y=gnHistogramOrgY+gnHistogramHeight-(int)(fy*gnHistogramHeight+0.5);
		MoveToEx( hdc, gnHistogramOrgX, y, NULL );
		LineTo( hdc, gnHistogramOrgX-5, y );
	}

	if (gbStatisticsInfo) {
		// calculate median and average
		total=width*height;
		for (c=0; c<3; c++) {
			acc=0;
			for (i=0; i<256; i++) {
				acc+=count[c][i];
				if (acc>total/2) {
					med[c]=i;
					break;
				}
			}
		}
		for (c=0; c<3; c++) {
			sum=0;
			for (i=0; i<256; i++) {
				sum+=count[c][i]*i;
			}
			avg[c]=(int)(sum/total+0.5);
		}
	}

	// accumulate count
	if (gnHistogramDelta!=1) {
		for (c=0; c<3; c++) {
			for (i=0; i<256; i+=gnHistogramDelta) {
				for (x=1; x<gnHistogramDelta; x++) {
					count[c][i/gnHistogramDelta]+=count[c][i/gnHistogramDelta+x];
				}
				if (count[c][i/gnHistogramDelta]>maxc[c])
					maxc[c]=count[c][i/gnHistogramDelta];
			}
		}
	}

	realmaxc=0;
	for (c=0; c<3; c++) {
		if (gbHistogramDisplay[c] && maxc[c]>realmaxc) realmaxc=maxc[c];
	}
	realmaxc=(int)(realmaxc*gdScaleFactor+0.5);

	if (!gbHistogramAutoNor) 
		realmaxc=(int)(gnHistogramMax*gdScaleFactor+0.5);

	// draw peak information
	if (gbStatisticsInfo && gnHistogramDelta==1) {
		SetTextColor(hdc, RGB(255, 255, 255));
		SetBkColor(hdc, RGB(0, 0, 0));
		hOldFont = (HFONT)SelectObject(hdc, ghPeakFont);

		TextOut(hdc, gnHistogramOrgX+5, gnHistogramOrgY+2, szBuffer, 
				wsprintf(szBuffer, "peak=%03d(%d),%03d(%d),%03d(%d)", 
				maxc_color[0], maxc[0], maxc_color[1], maxc[1], maxc_color[2], maxc[2]) );

		SelectObject(hdc, hOldFont);
	}

	if (realmaxc!=0) {

		// color blending
		SetROP2(hdc, R2_MERGEPEN);
		// draw it
		for (c=0; c<3; c++) {
			if (gbHistogramDisplay[c]==FALSE) continue;

			// prepare points
			for (i=0; i<256; i+=gnHistogramDelta) {
				ptHist[i/gnHistogramDelta].x=gnHistogramOrgX+i/gnHistogramDelta;
				ptHist[i/gnHistogramDelta].y=gnHistogramOrgY+gnHistogramHeight
												-(count[c][i]*gnHistogramHeight+(realmaxc/2))/realmaxc;
												// note that i doesn't have to be divided bt Delta
			}
			selectPen(hdc, c);
			Polyline(hdc, ptHist, 256/gnHistogramDelta);
		}
		SetROP2(hdc, R2_COPYPEN);

	}

	SelectObject(hdc, GetStockObject(WHITE_PEN));
	SelectObject(hdc, GetStockObject(NULL_BRUSH));
	Rectangle(hdc, grectHistogram.left, grectHistogram.top, grectHistogram.right, grectHistogram.bottom);

	SelectClipRgn(hdc, ghrgnCanvasClip);

	if (gbStatisticsInfo) {

		// print statistics 
//		gnHistogramTextY=100;
		SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
		SetTextColor(hdc, RGB(255, 255, 255));
		SetBkColor(hdc, RGB(0, 0, 0));
		if (gnHistogramDelta==1) {
			TextOut(hdc, gnHistogramTextX,     gnHistogramTextY+0, szBuffer,
						wsprintf(szBuffer, "min=(%03d,%03d,%03d)", min[0], min[1], min[2]) );
			TextOut(hdc, gnHistogramTextX+136, gnHistogramTextY+0, szBuffer,
						wsprintf(szBuffer, "max=(%03d,%03d,%03d)", max[0], max[1], max[2]) );
			TextOut(hdc, gnHistogramTextX,     gnHistogramTextY+TEXTBOX_HEIGHT, szBuffer,
						wsprintf(szBuffer, "avg=(%03d,%03d,%03d)", avg[0], avg[1], avg[2]) );
			TextOut(hdc, gnHistogramTextX+136, gnHistogramTextY+TEXTBOX_HEIGHT, szBuffer,
						wsprintf(szBuffer, "med=(%03d,%03d,%03d)", med[0], med[1], med[2]) );
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_2_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_2_ROWS);
		} else if (gnHistogramDelta==2) {
			TextOut(hdc, gnHistogramTextX, gnHistogramTextY+0, szBuffer,
						wsprintf(szBuffer, "min=(%03d,%03d,%03d)", min[0], min[1], min[2]) );
			TextOut(hdc, gnHistogramTextX, gnHistogramTextY+TEXTBOX_HEIGHT, szBuffer,
						wsprintf(szBuffer, "max=(%03d,%03d,%03d)", max[0], max[1], max[2]) );
			TextOut(hdc, gnHistogramTextX, gnHistogramTextY+TEXTBOX_HEIGHT*2, szBuffer,
						wsprintf(szBuffer, "avg=(%03d,%03d,%03d)", avg[0], avg[1], avg[2]) );
			TextOut(hdc, gnHistogramTextX, gnHistogramTextY+TEXTBOX_HEIGHT*3, szBuffer,
						wsprintf(szBuffer, "med=(%03d,%03d,%03d)", med[0], med[1], med[2]) );
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_1_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_4_ROWS);
		}
	}
/*
	ReleaseScreenCapture();
*/
}


VOID DoTheGradientGraph(HDC hdc, HDC hdcScreen)
{
	int i;
	int x, y;

	switch (gnRGBGraphMode) {
	case MENU_GRAPH_HOR_STEP:
	case MENU_GRAPH_HOR_LINEAR:
		// select clip region
		SelectClipRgn(hdc, ghrgnHorGraphClip);

		FillRect( hdc, &grectHorGraph, (HBRUSH)GetStockObject(BLACK_BRUSH));

		if (gbRGBGraphHighlight && gnColorBits>=24) {
			SelectObject(hdc, GetStockObject(NULL_PEN));
			SelectObject(hdc, GetStockObject(DKGRAY_BRUSH));
			Rectangle(hdc, gnZoom*(gcxZoomed/2)-1, gnHorGraphTop+1,
						   gnZoom*(gcxZoomed/2+1)+1, gnHorGraphBase-1);
		}

		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP) {
			//prepare points
			x = 0;
			for (i=0; i<gcxZoomed; i++) {
				gptBuf[0][i].x = x;
				gptBuf[0][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetRValue(gcolorBuf[i])) - 2;
				x += gnZoom;
			}

			SelectObject(hdc, GetStockObject(WHITE_PEN));
			for (i=0; i<gcxZoomed; i++) {
				MoveToEx(hdc, gptBuf[0][i].x,        gptBuf[0][i].y, NULL);
				LineTo  (hdc, gptBuf[0][i].x+gnZoom, gptBuf[0][i].y);
			}
		} else {
			//prepare points
			x = gnZoom/2;
			for (i=0; i<gcxZoomed; i++) {
				gptBuf[0][i].x = x;
				gptBuf[0][i].y = gnHorGraphBase - (int)(gfRGBHorFactor*GetRValue(gcolorBuf[i])) - 2;
				x += gnZoom;
			}

			SelectObject(hdc, GetStockObject(WHITE_PEN));
			Polyline(hdc, gptBuf[0], gcxZoomed);
		}

		// draw outside
		SelectObject(hdc, GetStockObject(WHITE_PEN));
		SelectObject(hdc, GetStockObject(NULL_BRUSH));
		Rectangle(hdc, grectHorGraph.left, grectHorGraph.top, grectHorGraph.right, grectHorGraph.bottom);
		break;

	case MENU_GRAPH_VER_STEP:
	case MENU_GRAPH_VER_LINEAR:
		// select clip region
		SelectClipRgn(hdc, ghrgnVerGraphClip);

		FillRect( hdc, &grectVerGraph, (HBRUSH)GetStockObject(BLACK_BRUSH));

		if (gbRGBGraphHighlight && gnColorBits>=24) {
			SelectObject(hdc, GetStockObject(NULL_PEN));
			SelectObject(hdc, GetStockObject(DKGRAY_BRUSH));
			Rectangle(hdc, gnVerGraphBase+1, gnZoom*(gcyZoomed/2)-1, 
						   gnVerGraphTop-1, gnZoom*(gcyZoomed/2+1)+1);
		}

		if (gnRGBGraphMode==MENU_GRAPH_VER_STEP) {
			//prepare points
			y = 0;
			for (i=0; i<gcyZoomed; i++) {
				gptBuf[0][i].y = y;
				gptBuf[0][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetRValue(gcolorBuf[i])) + 2;
				y += gnZoom;
			}

			SelectObject(hdc, GetStockObject(WHITE_PEN));
			for (i=0; i<gcyZoomed; i++) {
				MoveToEx(hdc, gptBuf[0][i].x, gptBuf[0][i].y, NULL);
				LineTo  (hdc, gptBuf[0][i].x, gptBuf[0][i].y+gnZoom);
			}
		} else {
			//prepare points
			y = gnZoom/2;
			for (i=0; i<gcyZoomed; i++) {
				gptBuf[0][i].y = y;
				gptBuf[0][i].x = gnVerGraphBase + (int)(gfRGBVerFactor*GetRValue(gcolorBuf[i])) + 2;
				y += gnZoom;
			}

			SelectObject(hdc, GetStockObject(WHITE_PEN));
			Polyline(hdc, gptBuf[0], gcyZoomed);
		}

		// draw outside
		SelectObject(hdc, GetStockObject(WHITE_PEN));
		SelectObject(hdc, GetStockObject(NULL_BRUSH));
		Rectangle(hdc, grectVerGraph.left, grectVerGraph.top, grectVerGraph.right, grectVerGraph.bottom);

		break;
	case MENU_GRAPH_STATISTICS:
		
		break;
	}

	SelectClipRgn(hdc, ghrgnCanvasClip);
}

float gfGradientMax, gfGradientMin, gfGradientAvg;

VOID DrawTheGradientStatistics(HDC hdc, HDC hdcScreen)
{
	int x, y;
	int i;
	int maxc;
	float fy;
	POINT ptHist[256];
	char szBuffer[128];

	// draw the diagram
	SelectClipRgn(hdc, ghrgnHistogram);

	FillRect( hdc, &grectHistogram, (HBRUSH)GetStockObject(BLACK_BRUSH));

	// draw the ruler
	SelectObject(hdc, GetStockObject(WHITE_PEN));
	// horizontal grid
	MoveToEx( hdc, gnHistogramOrgX-HISTOGRAM_RULER_HEIGHT, gnHistogramOrgY+gnHistogramHeight, NULL );
	LineTo  ( hdc, gnHistogramOrgX+gnHistogramWidth, gnHistogramOrgY+gnHistogramHeight );
	for (x=0; x<gnHistogramWidth; x+=gnHistogramGridX) {
		MoveToEx( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight, NULL );
		if ( (gnHistogramDelta==1 && (x%50)==0) || (gnHistogramDelta!=1 && x%25==0) ) 
			LineTo( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight+6 );
		else
			LineTo( hdc, gnHistogramOrgX+x, gnHistogramOrgY+gnHistogramHeight+4 );
	}
	// vertical grid
	MoveToEx( hdc, gnHistogramOrgX, gnHistogramOrgY, NULL );
	LineTo  ( hdc, gnHistogramOrgX, gnHistogramOrgY+gnHistogramHeight+HISTOGRAM_RULER_HEIGHT );
	for (fy=0; fy<=1.0; fy+=0.1f) {
		y=gnHistogramOrgY+gnHistogramHeight-(int)(fy*gnHistogramHeight+0.5);
		MoveToEx( hdc, gnHistogramOrgX, y, NULL );
		LineTo( hdc, gnHistogramOrgX-5, y );
	}

	// accumulate count
	if (gnHistogramDelta!=1) {
		for (i=0; i<256; i+=gnHistogramDelta) {
			for (x=1; x<gnHistogramDelta; x++) {
				count[0][i/gnHistogramDelta]+=count[0][i/gnHistogramDelta+x];
			}
		}
	}
	// find the max count
	maxc=0;
	for (x=1; x<256; x+=gnHistogramDelta) {
		if (count[0][x]>maxc) maxc=count[0][x];
	}
	maxc=(int)(maxc*gdScaleFactor+0.5);

	if (maxc!=0) {
		// draw it
		// prepare points
		for (i=0; i<256; i+=gnHistogramDelta) {
			ptHist[i/gnHistogramDelta].x=gnHistogramOrgX+i/gnHistogramDelta;
			ptHist[i/gnHistogramDelta].y=gnHistogramOrgY+gnHistogramHeight
											-(count[0][i]*gnHistogramHeight+(maxc/2))/maxc;
											// note that i doesn't have to be divided bt Delta
		}
		SelectObject(hdc, GetStockObject(WHITE_PEN));
		Polyline(hdc, ptHist, 256/gnHistogramDelta);
	}

	SelectObject(hdc, GetStockObject(WHITE_PEN));
	SelectObject(hdc, GetStockObject(NULL_BRUSH));
	Rectangle(hdc, grectHistogram.left, grectHistogram.top, grectHistogram.right, grectHistogram.bottom);

	SelectClipRgn(hdc, ghrgnCanvasClip);

	if (gbStatisticsInfo) {
		// print statistics 
//		gnHistogramTextY=100;
		SelectObject(hdc, GetStockObject(SYSTEM_FIXED_FONT));
		SetTextColor(hdc, RGB(255, 255, 255));
		SetBkColor(hdc, RGB(0, 0, 0));
		SelectObject(hdc, GetStockObject(BLACK_BRUSH));
		if (gnHistogramDelta==1) {
			SelectObject(hdc, GetStockObject(BLACK_BRUSH));
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_2_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_2_ROWS);

			TextOut(hdc, gnHistogramTextX+20,     gnHistogramTextY+0, szBuffer,
						sprintf(szBuffer, "min=%8.4f", gfGradientMin) );
			TextOut(hdc, gnHistogramTextX+136+20, gnHistogramTextY+0, szBuffer,
						sprintf(szBuffer, "max=%8.4f", gfGradientMax) );
			TextOut(hdc, gnHistogramTextX+20,     gnHistogramTextY+TEXTBOX_HEIGHT, szBuffer,
						sprintf(szBuffer, "avg=%8.4f", gfGradientAvg) );
//			TextOut(hdc, gnHistogramTextX+136, gnHistogramTextY+TEXTBOX_HEIGHT, szBuffer,
//						wsprintf(szBuffer, "med=(%03d,%03d,%03d)", med[0], med[1], med[2]) );

			SelectObject(hdc, GetStockObject(NULL_BRUSH));
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_2_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_2_ROWS);

		} else if (gnHistogramDelta==2) {
			SelectObject(hdc, GetStockObject(BLACK_BRUSH));
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_1_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_4_ROWS);

			TextOut(hdc, gnHistogramTextX+20, gnHistogramTextY+0+8, szBuffer,
						sprintf(szBuffer, "min=%8.4f", gfGradientMin) );
			TextOut(hdc, gnHistogramTextX+20, gnHistogramTextY+TEXTBOX_HEIGHT+8, szBuffer,
						sprintf(szBuffer, "max=%8.4f", gfGradientMax) );
			TextOut(hdc, gnHistogramTextX+20, gnHistogramTextY+TEXTBOX_HEIGHT*2+8, szBuffer,
						sprintf(szBuffer, "avg=%8.4f", gfGradientAvg) );
//			TextOut(hdc, gnHistogramTextX, gnHistogramTextY+TEXTBOX_HEIGHT*3, szBuffer,
//						wsprintf(szBuffer, "med=(%03d,%03d,%03d)", med[0], med[1], med[2]) );

			SelectObject(hdc, GetStockObject(NULL_BRUSH));
			Rectangle(hdc, gnHistogramTextX, gnHistogramTextY, 
							gnHistogramTextX+TEXTBOX_WIDTH_1_COLS, gnHistogramTextY+TEXTBOX_HEIGHT_4_ROWS);

		}
	}
}

// The speedup is not very successful
//#define SPEEDUP		1

VOID DoTheGradient(HDC hdc, HDC hdcScreen)
{
	int x, y;
	int cx, cy;
	int c;
	int r, g, b;
	int t1, dx, dy;
	float max;
	float min;
	float avg, temp;
#ifndef SPEEDUP
	int t2;
#endif

	unsigned char *gray, *pGray;
	float *gradient, *pGradient;
#ifdef SPEEDUP
	int i;
	int *accBuf[3];
	int *prev, *cur, *next;
	int *pPrev, *pCur, *pNext;
#endif
	unsigned char *pData;
	unsigned char *buf;

	int nMemoryWidth;
	int width, height;
	int padding;

	if (gnColorBits < 24) return;

	width=gcxZoomed+2;
	height=gcyZoomed+2;
	ScreenCapture(hdcScreen, gnx-gcxZoomed/2-1, gny-gcyZoomed/2-1, width, height,
				  &buf, &padding);
	nMemoryWidth = gnCaptureMemoryWidth;

	// gray image 
	gray = (unsigned char *)malloc(width*height);

	pData=buf;
	pGray=gray;
	for (cy=0; cy<height; cy++) {
		for (cx=0; cx<width; cx++) {

			b=*(pData++);
			g=*(pData++);
			r=*(pData++);

			// I=(2R+5G+B)/8
			c=((r+2)>>2)+((((g<<2)+g+4)>>3))+((b+4)>>3);
			if (c>255) *pGray=255;
			else *pGray=c;

			if (gnColorBits==32) pData++;
			pGray++;
		}
		pData+=padding;
	}

#ifdef SPEEDUP
	for (i=0; i<3; i++) {
		accBuf[i]=malloc(max(height, width)*sizeof(int));
	}

	max=0;
	gradient=malloc(width*height*sizeof(float));

	// compute ver. gradient
	prev=accBuf[0];
	cur=accBuf[1];
	next=accBuf[2];
	// precompute two rows
	pPrev=prev;
	pGray=gray+1;
	for (x=1; x<width-1; x++) {
		*pPrev=pGray[-1]+2*pGray[0]+pGray[1];
		pGray++;
		pPrev++;
	}
	pGray+=2;	// advance to the 2nd one of the next row 
	pCur=cur;
	for (x=1; x<width-1; x++) {
		*pCur=pGray[-1]+2*pGray[0]+pGray[1];
		pGray++;
		pCur++;
	}
	pGray+=2;

	pGradient=gradient+width+1;
	for (y=1; y<height-1; y++) {
		pCur=cur;
		pNext=next;
		for (x=1; x<width-1; x++) {
			// compute the next row
			*pNext=pGray[-1]+2*pGray[0]+pGray[1];
			// compute gradient
			dx=(*pNext)-(*pCur);
			*pGradient=(float)(dx*dx);

			//if (*pGradient>max) max=*pGradient;

			pGradient++;
			pGray++;
			pNext++;
			pCur++;
		}
		pGray+=2;
		pGradient+=2;

		prev=cur;
		cur=next;
		next=prev;
	}

		// compute ver. gradient
	prev=accBuf[0];
	cur=accBuf[1];
	next=accBuf[2];
	// precompute two rows
	pPrev=prev;
	pGray=gray+width;
	for (y=1; y<height-1; y++) {
		*pPrev=pGray[-width]+2*pGray[0]+pGray[width];
		pGray+=width;
		pPrev++;
	}
	pGray=gray+width+1;	// advance to the 2nd one of the next row 
	pCur=cur;
	for (y=1; y<height-1; y++) {
		*pCur=pGray[-width]+2*pGray[0]+pGray[width];
		pGray+=width;
		pCur++;
	}

	for (x=1; x<width-1; x++) {
		pGradient=gradient+width+x;
		pGray=gray+width+x;
		pCur=cur;
		pNext=next;
		for (y=1; y<height-1; y++) {
			// compute the next row
			*pNext=pGray[-width]+2*pGray[0]+pGray[width];
			// compute gradient
			dy=(*pNext)-(*pCur);
			*pGradient=(float)sqrt((*pGradient)+(dy*dy))/4;

			if (*pGradient>max) max=*pGradient;

			pGradient+=width;
			pGray+=width;
			pNext++;
			pCur++;
		}

		prev=cur;
		cur=next;
		next=prev;
	}
#endif

#ifndef SPEEDUP
#define GRAY(x, y)	(gray[(y)*width+(x)])

	max=0;
	min=10000;
	avg=0;
	gradient=(float *)malloc(width*height*sizeof(float));
	pGradient=gradient+width;
	for (y=1; y<height-1; y++) {
		pGradient++;
		temp=0;
		for (x=1; x<width-1; x++) {
			t1=GRAY(x+1, y-1)+2*GRAY(x+1, y)+GRAY(x+1, y+1);
			t2=GRAY(x-1, y-1)+2*GRAY(x-1, y)+GRAY(x-1, y+1);
			dx=t1-t2;

			t1=GRAY(x-1, y-1)+2*GRAY(x, y-1)+GRAY(x+1, y-1);
			t2=GRAY(x-1, y+1)+2*GRAY(x, y+1)+GRAY(x+1, y+1);
			dy=t1-t2;

			*pGradient=(float)sqrt((float)dx*(float)dx+(float)dy*(float)dy)/4;

			if (*pGradient>max) max=*pGradient;

			if (gbStatisticsInfo) {
				if (*pGradient<min) min=*pGradient;
				temp+=*pGradient;
			}

			pGradient++;
		}
		pGradient++;
		if (gbStatisticsInfo) {
			avg+=temp/width;
		}
	}
	if (gbStatisticsInfo) {
		avg=avg/height;

		gfGradientMax=max;
		gfGradientMin=min;
		gfGradientAvg=avg;
	}

	if (gbGradientAutoNor) {
		// adjust brightness
		max=max*gdGradientScaleFactor;
	} else {
		max=(float)gnGradientMax*gdGradientScaleFactor; // *gdGradientScaleFactor?
	}

#undef GRAY
#endif

	// reset count if necessary
	if (gnRGBGraphMode == MENU_GRAPH_STATISTICS)
		for (c=0; c<256; c++) count[0][c]=0;

	// make gradient image displayable 
	pData=buf+nMemoryWidth+(gnColorBits/8);
	pGradient=gradient+width+1;
	for (cy=1; cy<height-1; cy++) {
		for (cx=1; cx<width-1; cx++) {

			t1=(int)(*pGradient/max*255);
			if (t1>255) t1=255;

			t1=ContrastLUT[BrightnessLUT[GammaLUT[t1]]];

			*(pData++)=t1;
			*(pData++)=t1;
			*(pData++)=t1;

			if (gnRGBGraphMode == MENU_GRAPH_STATISTICS)
				count[0][t1]++;

			if (gnColorBits==32) pData++;
			pGradient++;
		}
		pData+=padding+2*(gnColorBits/8);
		pGradient+=2;
	}

	// draw highlight if necessary
	if (gbRGBGraphHighlight) {
		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
			pData=buf+nMemoryWidth*(height-1-height/2)+(gnColorBits/8);
			for (cx=1; cx<width; cx++) {
				t1=*pData+30;
				if (t1>255) t1=255;

				pData++;
				*(pData++)=t1;
				*(pData++)=t1;	
//				*(pData++)=t1;

				if (gnColorBits==32) pData++;
			}
		} else
		if (gnRGBGraphMode==MENU_GRAPH_VER_STEP || gnRGBGraphMode==MENU_GRAPH_VER_LINEAR) {
			pData=buf+nMemoryWidth+(width/2)*(gnColorBits/8);
			for (cy=1; cy<height-1; cy++) {
				t1=*pData+30;
				if (t1>255) t1=255;

//				pData[0]=t1;
				pData[1]=t1;
				pData[2]=t1;

				pData+=nMemoryWidth;
			}
		}
	}

	// copy to hdc
//	StretchBlt(hdc, 0, 0, (width-2)*gnZoom, (height-2)*gnZoom, hdcMem, 1, 1, width-2, height-2, SRCCOPY);
	StretchBlt(hdc, 0, 0, (width-2)*gnZoom, (height-2)*gnZoom, hdcCaptureMem, 1, 1, width-2, height-2, SRCCOPY);

	// fill in color buffer if drawing graph is necessary
	if (gnRGBGraphMode != MENU_GRAPH_OFF && gnRGBGraphMode != MENU_GRAPH_STATISTICS) {
		int i;
		// note that y is from bottom to top as bmp order
		if (gnRGBGraphMode==MENU_GRAPH_HOR_STEP || gnRGBGraphMode==MENU_GRAPH_HOR_LINEAR) {
//			pGradient=gradient+width*(height-1-height/2)+1;
			pData=buf+nMemoryWidth*(height-1-height/2)+(gnColorBits/8);
			for (i=0; i<width-2; i++) {
//				t1=(int)(*pGradient/max*255);
				t1=*pData;
				if (t1>255) t1=255;
				gcolorBuf[i]=t1;
//				pGradient++;
				if (gnColorBits==32) pData+=4;
				else pData+=3;
			}
		} else {
//			pGradient=gradient+(height-2)*width+(width/2);
			pData=buf+nMemoryWidth*(height-2)+(gnColorBits/8)*(width/2);
			for (i=0; i<height-2; i++) {
//				t1=(int)(*pGradient/max*255);
				t1=*pData;
				if (t1>255) t1=255;
				gcolorBuf[i]=t1;
//				pGradient-=width;
				pData-=nMemoryWidth;
			}
		}	
	}

#ifdef SPEEDUP
	for (i=0; i<3; i++) free(accBuf[i]);
#endif
	free(gray);
	free(gradient);

	ReleaseScreenCapture();
}

/************************************************************************
* DoTheZoomIn
*
* Does the actual paint of the zoomed image.
*
* Arguments:
*   HDC hdc - If not NULL, this hdc will be used to paint with.
*             If NULL, a dc for the apps window will be obtained.
*
* History:
*
************************************************************************/

VOID DoTheZoomIn(
    HDC hdc)
{
    BOOL fRelease;
    HPALETTE hpalOld = NULL;
    HDC hdcScreen;

    HPALETTE hpalMemOld = NULL;
    HDC hdcMem;
    HBITMAP hBmpMem;
    HBITMAP hBmpMemOld;

	unsigned char *buf;
	int padding;
	BOOL fCapture=FALSE;

	if (gbQuit) return;

	if (gbInit) return;

    if (!hdc) {
        hdc = GetDC(ghwndApp);
        fRelease = TRUE;
    }
    else {
        fRelease = FALSE;
    }

    // I use an off-screen dc to avoid flickering
    hdcMem = CreateCompatibleDC(hdc);
	if (gbRuler)
		hBmpMem = CreateCompatibleBitmap(hdc, gnZoom*gcxZoomed+RULER_HEIGHT, gnZoom*gcyZoomed+RULER_HEIGHT);
	else
		hBmpMem = CreateCompatibleBitmap(hdc, gnZoom*gcxZoomed, gnZoom*gcyZoomed);
    hBmpMemOld = (HBITMAP)SelectObject(hdcMem, hBmpMem);

	if (gbRuler) {
		// draw the ruler
		DoTheRuler(hdcMem);
		// change the origin
		SetWindowOrgEx(hdcMem, -RULER_HEIGHT, -RULER_HEIGHT, NULL);
	}

	SelectClipRgn(hdcMem, ghrgnCanvasClip);

    if (ghpalPhysical) {
		hpalOld = SelectPalette(hdc, ghpalPhysical, FALSE);
        RealizePalette(hdc);

		hpalMemOld = SelectPalette(hdcMem, ghpalPhysical, FALSE);
        RealizePalette(hdcMem);
    }

    /*
     * The point must not include areas outside the screen dimensions.
     */
	if (gbBound) {
		if (gnDisplayMode==MENU_FILTER_SNOOP) {
			gnx = BOUND(gptZoom.x, gcxZoomed / 2, gcxScreenMax - (gcxZoomed / 2));
			gny = BOUND(gptZoom.y, gcyZoomed / 2, gcyScreenMax - (gcyZoomed / 2));
		} else {
			gnx = BOUND(gptZoom.x, gcxZoomed/2+1, gcxScreenMax - (gcxZoomed/2) -1);
			gny = BOUND(gptZoom.y, gcyZoomed/2+1, gcyScreenMax - (gcyZoomed/2) -1);
		}
	} else {
		gnx = gptZoom.x;
		gny = gptZoom.y;
	}

	hdcScreen = GetDC(NULL);

	switch (gnDisplayMode) {
		case MENU_FILTER_SNOOP:
		    // Draw in the off-screen dc
			if (gnColorBits<24) {
			    SetStretchBltMode(hdcMem, COLORONCOLOR);
			    StretchBlt(hdcMem, 0, 0, gnZoom * gcxZoomed, gnZoom * gcyZoomed,
					    hdcScreen, gnx - gcxZoomed / 2,
				        gny - gcyZoomed / 2, gcxZoomed, gcyZoomed, SRCCOPY);
				// delayed the release of hdcScreen to exit
			} else {
				fCapture=TRUE;
				ScreenCapture(hdcScreen, gnx - gcxZoomed / 2, gny - gcyZoomed / 2, 
								gcxZoomed, gcyZoomed, &buf, &padding);

				// make the adjustment if necessary
				if (gnBrightness!=0 || gnContrast!=0 || gfGamma!=1.0f) {
					AdjustColor(buf, gcxZoomed, gcyZoomed, padding);
				}

			    SetStretchBltMode(hdcMem, COLORONCOLOR);
			    StretchBlt(hdcMem, 0, 0, gnZoom * gcxZoomed, gnZoom * gcyZoomed,
					    hdcCaptureMem, 0, 0, gcxZoomed, gcyZoomed, SRCCOPY);
			}

			// draw highlight first if necessary, it must be drawn before we draw grid 
		    //   and the center highlight and after we stretch the capture
			if (gnRGBGraphMode != MENU_GRAPH_OFF && gnRGBGraphMode != MENU_GRAPH_STATISTICS) {
				// fill in color buffer
 				// we use the off-screen dc already drawn to avoid causing lag in Windows Vista Aero environments
 				FillinColorBuffer(hdcMem);  // patched by mifan 2009.1.10
 				// FillinColorBuffer(hdcScreen);

				// draw highlight if necessary
				if (gbRGBGraphHighlight)
					DoTheHighlight(hdcMem, hdcScreen);
			}

		    if (gbGrid)
				DoTheGrid(hdcMem, GetStockObject(BLACK_PEN));

			/*
			 * do all the other rendering jobs here
			 */
			// draw RGB graph (must be drawn after drawing grid)
			if (gnRGBGraphMode != MENU_GRAPH_OFF) {
				if (gnRGBGraphMode != MENU_GRAPH_STATISTICS) 
					DoTheRGBGraph(hdcMem, hdcScreen);
				else {
					// for statistics, the scrren has been captured already
					// don't need to do that again in DoTheStatistics
					DoTheStatistics(hdcMem, hdcScreen);
				}
			}

			if (gbCenterHighlight)
				DoTheCenterHightlight(hdcMem);

		    /*
		     * and draw in the windows dc
			*/

			if (fCapture)
				ReleaseScreenCapture();

		break;
		case MENU_FILTER_GRADIENT:
			// draw gradient
			DoTheGradient(hdcMem, hdcScreen);

		    if (gbGrid)
				DoTheGrid(hdcMem, ghGPen);

			if (gnRGBGraphMode != MENU_GRAPH_OFF) {
				if (gnRGBGraphMode != MENU_GRAPH_STATISTICS) 
					DoTheGradientGraph(hdcMem, hdcScreen);
				else 
					DrawTheGradientStatistics(hdcMem, hdcScreen);
			}

			if (gbCenterHighlight)
				DoTheCenterHightlight(hdcMem);
		break;
	}		

	// draw status bar
	{
		POINT  p;
		RECT   r;
		char buf[128];
		COLORREF rgb=GetPixel(hdcScreen, gnx, gny);

		if (gnCoordSystem==MENU_ACTIVE_WINDOW) {
			if (!gbInverseY) {
				p.x=gnx;	p.y=gny;
				ScreenToClient(GetForegroundWindow(), &p);
			}
			else {
				p.x=gnx;	p.y=gny;
				ScreenToClient(GetForegroundWindow(), &p);
				GetClientRect(GetForegroundWindow(), &r);
				p.y=(r.bottom-r.top+1)-p.y;
			}

			if (gbDisplayHex)
				sprintf(buf, "(%04d,%04d) - (%02X,%02X,%02X) W", p.x, p.y,
					GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
			else
				sprintf(buf, "(%04d,%04d) - (%03d,%03d,%03d) W", p.x, p.y,
					GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
		} else {
			if (!gbInverseY) {
				if (gbDisplayHex)
					sprintf(buf, "(%04d,%04d) - (%02X,%02X,%02X) S", gnx, gny,
						GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
				else
					sprintf(buf, "(%04d,%04d) - (%03d,%03d,%03d) S", gnx, gny,
						GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
			} else {
				if (gbDisplayHex)
					sprintf(buf, "(%04d,%04d) - (%02X,%02X,%02X) S", gnx, gnWindowsHeight-gny,
						GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
				else
					sprintf(buf, "(%04d,%04d) - (%03d,%03d,%03d) S", gnx, gnWindowsHeight-gny,
						GetRValue(rgb), GetGValue(rgb), GetBValue(rgb));
			}
		}

		SendMessage(ghwndStatusBar, SB_SETTEXT, 0, (LPARAM)buf);
	}


	// copy in back buffer
	if (gbRuler) {
		// change the origin back
		SetWindowOrgEx(hdcMem, 0, 0, NULL);
	}

	SetStretchBltMode(hdc, COLORONCOLOR);
	BitBlt(hdc, 0, 0, gnClientWidth, gnClientHeight-SB_HEIGHT, hdcMem, 0, 0, SRCCOPY);

	if (hpalOld) {
		SelectPalette(hdc, hpalOld, FALSE);
		SelectPalette(hdcMem, hpalMemOld, FALSE);
	}

    SelectObject(hdcMem, hBmpMemOld);
    DeleteObject(hBmpMem);
    DeleteDC(hdcMem);

    if (fRelease)
        ReleaseDC(ghwndApp, hdc);


	ReleaseDC(NULL, hdcScreen);
}



/************************************************************************
* 
*
* This function moves the current view around.
*
* Arguments:
*   INT nDirectionCode - Direction to move.  Must be VK_UP, VK_DOWN,
*                        VK_LEFT or VK_RIGHT.
*   BOOL fFast         - TRUE if the move should jump a larger increment.
*                        If FALSE, the move is just one pixel.
*   BOOL fPeg          - If TRUE, the view will be pegged to the screen
*                        boundary in the specified direction.  This overides
*                        the fFast parameter.
*
* History:
*
************************************************************************/

VOID MoveView(
    INT nDirectionCode,
    BOOL fFast,
    BOOL fPeg)
{
    INT delta;

    if (fFast)
        delta = gnFastDelta;
    else
        delta = 1;

    switch (nDirectionCode) {
        case VK_UP:
            if (fPeg)
                gptZoom.y = gcyZoomed / 2;
            else
                gptZoom.y -= delta;

			if (gbBound)
				gptZoom.y = BOUND(gptZoom.y, 0, gcyScreenMax);

            break;

        case VK_DOWN:
            if (fPeg)
                gptZoom.y = gcyScreenMax - (gcyZoomed / 2);
            else
                gptZoom.y += delta;

			if (gbBound)
				gptZoom.y = BOUND(gptZoom.y, 0, gcyScreenMax);

            break;

        case VK_LEFT:
            if (fPeg)
                gptZoom.x = gcxZoomed / 2;
            else
                gptZoom.x -= delta;

			if (gbBound)
				gptZoom.x = BOUND(gptZoom.x, 0, gcxScreenMax);

            break;

        case VK_RIGHT:
            if (fPeg)
                gptZoom.x = gcxScreenMax - (gcxZoomed / 2);
            else
                gptZoom.x += delta;

			if (gbBound)
				gptZoom.x = BOUND(gptZoom.x, 0, gcxScreenMax);

            break;
    }

    DoTheZoomIn(NULL);
}

VOID ResizeWindow(
    INT nDirectionCode,
    BOOL fFast)
{
	RECT rect;
    INT delta;
	int w, h;

    if (fFast)
        delta = gnFastDelta;
    else
        delta = 1;

	GetWindowRect(ghwndApp, &rect);

    switch (nDirectionCode) {
        case VK_UP:
			w = rect.right-rect.left;

			h = (rect.bottom-rect.top)-delta*gnZoom;
			BOUND(h, 40+gnZoom, gcyScreenMax/gnZoom);
			if (rect.top+h>gcyScreenMax) h=gcyScreenMax-rect.top;
            break;

        case VK_DOWN:
			w = rect.right-rect.left;

			h = (rect.bottom-rect.top)+delta*gnZoom;
			BOUND(h, 40+gnZoom, gcyScreenMax/gnZoom);
			if (rect.top+h>gcyScreenMax) h=gcyScreenMax-rect.top;
            break;

        case VK_LEFT:
			h = rect.bottom-rect.top;

			w = (rect.right-rect.left)-delta*gnZoom;
			BOUND(w, gnZoom, gcxScreenMax/gnZoom);
			if (rect.left+w>gcxScreenMax) w=gcxScreenMax-rect.left;
            break;

        case VK_RIGHT:
			h = rect.bottom-rect.top;

			w = (rect.right-rect.left)+delta*gnZoom;
			BOUND(w, gnZoom, gcxScreenMax/gnZoom);
			if (rect.left+w>gcxScreenMax) w=gcxScreenMax-rect.left;
            break;
    }

	SetWindowPos(ghwndApp, (gbOnTop ? HWND_TOPMOST : HWND_NOTOPMOST), 0,0,
					w,h, SWP_NOMOVE);

//    DoTheZoomIn(NULL);
}


/************************************************************************
* DrawZoomRect
*
* This function draws the tracking rectangle.  The size and shape of
* the rectangle will be proportional to the size and shape of the
* app's client, and will be affected by the zoom factor as well.
*
* History:
*
************************************************************************/

VOID DrawZoomRect(VOID)
{
    HDC hdc;
    RECT rc;
    INT x;
    INT y;

	if (gbBound) {
	    x = BOUND(gptZoom.x, gcxZoomed / 2, gcxScreenMax - (gcxZoomed / 2));
		y = BOUND(gptZoom.y, gcyZoomed / 2, gcyScreenMax - (gcyZoomed / 2));
	} else {
	    x = gptZoom.x;
		y = gptZoom.y;
	}

    rc.left   = x - gcxZoomed / 2;
    rc.top    = y - gcyZoomed / 2;
    rc.right  = rc.left + gcxZoomed;
    rc.bottom = rc.top + gcyZoomed;

    InflateRect(&rc, 1, 1);

    hdc = GetDC(NULL);

    PatBlt(hdc, rc.left,    rc.top,     rc.right-rc.left, 1,  DSTINVERT);
    PatBlt(hdc, rc.left,    rc.bottom,  1, -(rc.bottom-rc.top),   DSTINVERT);
    PatBlt(hdc, rc.right-1, rc.top,     1,   rc.bottom-rc.top,   DSTINVERT);
    PatBlt(hdc, rc.right,   rc.bottom-1, -(rc.right-rc.left), 1, DSTINVERT);

    ReleaseDC(NULL, hdc);
}



/************************************************************************
* EnableRefresh
*
* This function turns on or off the auto-refresh feature.
*
* Arguments:
*   BOOL fEnable - TRUE to turn the refresh feature on, FALSE to
*                  turn it off.
*
* History:
*
************************************************************************/

VOID EnableRefresh(
    BOOL fEnable)
{
    if (fEnable) {
        /*
         * Already enabled.  Do nothing.
         */
        if (gfRefEnable)
            return;

        if (SetTimer(ghwndApp, IDTIMER_SNOOP, gnRefInterval * 10, NULL))
            gfRefEnable = TRUE;
    }
    else {
        /*
         * Not enabled yet.  Do nothing.
         */
        if (!gfRefEnable)
            return;

        KillTimer(ghwndApp, IDTIMER_SNOOP);
        gfRefEnable = FALSE;
    }
}



/************************************************************************
* CopyToClipboard
*
* This function copies the client area image of the app into the
* clipboard.
*
* History:
*
************************************************************************/

VOID CopyToClipboard(VOID)
{
    HDC hdcSrc;
    HDC hdcDst;
    RECT rc;
    HBITMAP hbm;
	int w, h;
	int x, y;

    if (OpenClipboard(ghwndApp)) {
        EmptyClipboard();

        if (hdcSrc = GetDC(ghwndApp)) {
            GetClientRect(ghwndApp, &rc);
			w = rc.right - rc.left;
			h = rc.bottom - SB_HEIGHT - rc.top;
			if (gbRuler) {
				w = w - RULER_HEIGHT;
				h = h - RULER_HEIGHT;
			}
			x = rc.left;
			y = rc.top;
			if (gbRuler) {
				x = x + RULER_HEIGHT;
				y = y + RULER_HEIGHT;
			}
            if (hbm = CreateCompatibleBitmap(hdcSrc, w, h)) {
                if (hdcDst = CreateCompatibleDC(hdcSrc)) {
                    /*
                     * Calculate the dimensions of the bitmap and
                     * convert them to tenths of a millimeter for
                     * setting the size with the SetBitmapDimensionEx
                     * call.  This allows programs like WinWord to
                     * retrieve the bitmap and know what size to
                     * display it as.
                     */
                    SetBitmapDimensionEx(hbm, 
						(DWORD)((INT)(((DWORD)(w) * MM10PERINCH) / (DWORD)GetDeviceCaps(hdcSrc, LOGPIXELSX))), 
						(DWORD)((INT)(((DWORD)(h) * MM10PERINCH) / (DWORD)GetDeviceCaps(hdcSrc, LOGPIXELSY))), NULL);

                    SelectObject(hdcDst, hbm);
                    BitBlt(hdcDst, 0, 0,
                            w, h, hdcSrc, x, y, SRCCOPY);
                    DeleteDC(hdcDst);
                    SetClipboardData(CF_BITMAP, hbm);
                }
                else {
                    DeleteObject(hbm);
                }
            }

            ReleaseDC(ghwndApp, hdcSrc);
        }

        CloseClipboard();
    }
    else {
        MessageBeep(0);
    }
}


/************************************************************************
* AboutDlgProc
*
* This is the About Box dialog procedure.
*
* History:
*
************************************************************************/

BOOL APIENTRY AboutDlgProc(
    HWND hwnd,
    UINT msg,
    WPARAM wParam,
    LPARAM lParam)
{
    switch (msg) {
        case WM_INITDIALOG:
            return TRUE;

        case WM_COMMAND:
            EndDialog(hwnd, IDOK);
            return TRUE;

        default:
            return FALSE;
    }
}



/************************************************************************
* RefreshRateDlgProc
*
* This is the Refresh Rate dialog procedure.
*
* History:
*
************************************************************************/

BOOL APIENTRY RefreshRateDlgProc(
    HWND hwnd,
    UINT msg,
    WPARAM wParam,
    LPARAM lParam)
{
    BOOL fTranslated;

    switch (msg) {
        case WM_INITDIALOG:
            SendDlgItemMessage(hwnd, DID_REFRESHRATEINTERVAL, EM_LIMITTEXT,
                    3, 0L);
            SetDlgItemInt(hwnd, DID_REFRESHRATEINTERVAL, gnRefInterval, FALSE);
            CheckDlgButton(hwnd, DID_REFRESHRATEENABLE, gfRefEnable ? 1 : 0);
            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                case IDOK:
                    gnRefInterval = GetDlgItemInt(hwnd, DID_REFRESHRATEINTERVAL,
                            &fTranslated, FALSE);

                    /*
                     * Stop any existing timers then start one with the
                     * new interval if requested to.
                     */
                    EnableRefresh(FALSE);
                    EnableRefresh(
                            IsDlgButtonChecked(hwnd, DID_REFRESHRATEENABLE));

                    EndDialog(hwnd, IDOK);
                    break;

                case IDCANCEL:
                    EndDialog(hwnd, IDCANCEL);
                    break;
            }

            break;
    }

    return FALSE;
}

VOID SetSliderRange(HWND hSlider, int min, int max, BOOL bRedraw)
{
	SendMessage(hSlider, TBM_SETRANGEMIN, (WPARAM) FALSE, (LPARAM) min);
	SendMessage(hSlider, TBM_SETRANGEMAX, (WPARAM) bRedraw, (LPARAM) max);
}

VOID SetSliderPos(HWND hSlider, int pos, BOOL bRedraw)
{
	SendMessage(hSlider, TBM_SETPOS, (WPARAM) bRedraw, (LPARAM) pos);
}

INT GetSliderPos(HWND hSlider)
{
	return SendMessage(hSlider, TBM_GETPOS, (WPARAM) 0, (LPARAM) 0);
}

VOID SetSliderPageSize(HWND hSlider, int size)
{
	SendMessage(hSlider, TBM_SETPAGESIZE, (WPARAM) 0, (LPARAM) size);
}

#define SLIDER_OFFSET		100
#define SLIDER_MIN			0
#define SLIDER_MAX			199

BOOL APIENTRY HistogramDlgProc(
    HWND hwnd,
    UINT msg,
    WPARAM wParam,
    LPARAM lParam)
{
    BOOL fTranslated;
	INT pos;
	char buf[16];

	HWND hCtrl;
	static HWND hwndScrollBrightness;
	static HWND hwndScrollContrast;	
	static HWND hwndScrollGamma;

    switch (msg) {
        case WM_INITDIALOG:
            CheckDlgButton(hwnd, DID_HISTOGRAM_R, gbHistogramDisplay[0] ? 1 : 0);
            CheckDlgButton(hwnd, DID_HISTOGRAM_G, gbHistogramDisplay[1] ? 1 : 0);
            CheckDlgButton(hwnd, DID_HISTOGRAM_B, gbHistogramDisplay[2] ? 1 : 0);

            CheckDlgButton(hwnd, DID_HISTOGRAM_AUTO_NOR, gbHistogramAutoNor ? 1 : 0);
            CheckDlgButton(hwnd, DID_GRADIENT_AUTO_NOR, gbGradientAutoNor ? 1 : 0);

//            SendDlgItemMessage(hwnd, DID_HISTOGRAM_REF, EM_LIMITTEXT, 5, 0L);
            SetDlgItemInt(hwnd, DID_HISTOGRAM_REF, gnHistogramMax, FALSE);
            SetDlgItemInt(hwnd, DID_GRADIENT_REF, gnGradientMax, FALSE);

			hwndScrollBrightness=GetDlgItem(hwnd, DID_SLIDER_BRIGHTNESS);
			SetSliderRange(hwndScrollBrightness, SLIDER_MIN, SLIDER_MAX, FALSE);
			SetSliderPageSize(hwndScrollBrightness, 10);
			SetSliderPos(hwndScrollBrightness, gnBrightness+SLIDER_OFFSET, TRUE);
			SetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, gnBrightness, TRUE);

			hwndScrollContrast=GetDlgItem(hwnd, DID_SLIDER_CONTRAST);
			SetSliderRange(hwndScrollContrast, SLIDER_MIN, SLIDER_MAX, FALSE);
			SetSliderPageSize(hwndScrollContrast, 10);
			SetSliderPos(hwndScrollContrast, gnContrast+SLIDER_OFFSET, TRUE);
			SetDlgItemInt(hwnd, DID_EDIT_CONTRAST, gnContrast, TRUE);

			hwndScrollGamma=GetDlgItem(hwnd, DID_SLIDER_GAMMA);
			SetSliderRange(hwndScrollGamma, SLIDER_MIN, SLIDER_MAX, FALSE);
			SetSliderPageSize(hwndScrollGamma, 10);
			SetSliderPos(hwndScrollGamma, (int)(gfGamma*100), TRUE);
			sprintf(buf, "%1.2f", gfGamma);
			SetDlgItemText(hwnd, DID_EDIT_GAMMA, buf);

            return TRUE;

        case WM_COMMAND:
            switch (LOWORD(wParam)) {
                case IDOK:
				case DID_UPDATE:
                    gbHistogramDisplay[0] = (IsDlgButtonChecked(hwnd, DID_HISTOGRAM_R));
                    gbHistogramDisplay[1] = (IsDlgButtonChecked(hwnd, DID_HISTOGRAM_G));
                    gbHistogramDisplay[2] = (IsDlgButtonChecked(hwnd, DID_HISTOGRAM_B));

                    gbHistogramAutoNor = (IsDlgButtonChecked(hwnd, DID_HISTOGRAM_AUTO_NOR));
                    gbGradientAutoNor = (IsDlgButtonChecked(hwnd, DID_GRADIENT_AUTO_NOR));

                    gnHistogramMax = GetDlgItemInt(hwnd, DID_HISTOGRAM_REF, &fTranslated, FALSE);
                    gnGradientMax = GetDlgItemInt(hwnd, DID_GRADIENT_REF, &fTranslated, FALSE);

                    gnBrightness = GetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, &fTranslated, FALSE);
					if (gnBrightness>100) gnBrightness=100;
					else if (gnBrightness<-100) gnBrightness=-100;
					SetSliderPos(hwndScrollBrightness, gnBrightness+SLIDER_OFFSET, TRUE);
					SetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, gnBrightness, TRUE);
					SetBrightnessLUT(gnBrightness);

                    gnContrast = GetDlgItemInt(hwnd, DID_EDIT_CONTRAST, &fTranslated, FALSE);
					if (gnContrast>100) gnContrast=100;
					else if (gnContrast<-100) gnContrast=-100;
					SetSliderPos(hwndScrollContrast, gnContrast+SLIDER_OFFSET, TRUE);
					SetDlgItemInt(hwnd, DID_EDIT_CONTRAST, gnContrast, TRUE);
					SetContrastLUT(gnContrast);

					GetDlgItemText(hwnd, DID_EDIT_GAMMA, buf, 12);
					gfGamma=(float)atof(buf);
					if (gfGamma>2.0) gfGamma=2.0f;
					else if (gfGamma<0.02) gfGamma=0.02f;
					sprintf(buf, "%1.2f", gfGamma);
					gfGamma=(float)atof(buf);
					SetDlgItemText(hwnd, DID_EDIT_GAMMA, buf);
					SetGammaLUT(gfGamma);
					SetSliderPos(hwndScrollGamma, (int)(gfGamma*100), TRUE);

					if (LOWORD(wParam)==IDOK)
						EndDialog(hwnd, IDOK);
					else if (LOWORD(wParam)==DID_UPDATE)
						DoTheZoomIn(NULL);

                    break;

				case DID_RESET:
					gnBrightness = 0;
					SetSliderPos(hwndScrollBrightness, gnBrightness+SLIDER_OFFSET, TRUE);
					SetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, gnBrightness, TRUE);
					SetBrightnessLUT(gnBrightness);

                    gnContrast = 0;
					SetSliderPos(hwndScrollContrast, gnContrast+SLIDER_OFFSET, TRUE);
					SetDlgItemInt(hwnd, DID_EDIT_CONTRAST, gnContrast, TRUE);
					SetContrastLUT(gnContrast);

					gfGamma=1.0f;
					sprintf(buf, "%1.2f", gfGamma);
					SetDlgItemText(hwnd, DID_EDIT_GAMMA, buf);
					SetGammaLUT(gfGamma);
					SetSliderPos(hwndScrollGamma, (int)(gfGamma*100), TRUE);

					DoTheZoomIn(NULL);

					break;

                case IDCANCEL:
                    EndDialog(hwnd, IDCANCEL);
                    break;
            }

            break;

		case WM_HSCROLL: 
			hCtrl = (HWND)lParam;

            switch (LOWORD(wParam)) {
			case SB_PAGEDOWN:
			case SB_PAGEUP:
				if (hCtrl==hwndScrollBrightness) {
					gnBrightness=GetSliderPos(hCtrl)-SLIDER_OFFSET;
					SetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, gnBrightness, TRUE);
					SetBrightnessLUT(gnBrightness);
				} else if (hCtrl==hwndScrollContrast) {
					gnContrast=GetSliderPos(hCtrl)-SLIDER_OFFSET;
					SetDlgItemInt(hwnd, DID_EDIT_CONTRAST, gnContrast, TRUE);
					SetContrastLUT(gnContrast);
				} else if (hCtrl==hwndScrollGamma) {
					pos=GetSliderPos(hCtrl);
					sprintf(buf, "%1.2f", (float)pos/100);
					SetDlgItemText(hwnd, DID_EDIT_GAMMA, buf);
					SetGammaLUT(gfGamma=pos/100.0f);
				} 
				break;
			case SB_THUMBTRACK:
				if (hCtrl == hwndScrollBrightness) {
					gnBrightness=(short)HIWORD(wParam)-SLIDER_OFFSET;
					SetDlgItemInt(hwnd, DID_EDIT_BRIGHTNESS, gnBrightness, TRUE );
					SetBrightnessLUT(gnBrightness);
				} else if (hCtrl == hwndScrollContrast) {
					gnContrast=(short)HIWORD(wParam)-SLIDER_OFFSET;
					SetDlgItemInt(hwnd, DID_EDIT_CONTRAST, gnContrast, TRUE);
					SetContrastLUT(gnContrast);
				} else if (hCtrl == hwndScrollGamma) {
					pos=(short)HIWORD(wParam);
					sprintf(buf, "%1.2f", (float)pos/100);
					SetDlgItemText(hwnd, DID_EDIT_GAMMA, buf);
					SetGammaLUT(gfGamma=pos/100.0f);
				}
				break;
			}

			// necessary ?
			DoTheZoomIn(NULL);

			break;
    }

    return FALSE;
}

/****************************** Module Header *******************************
* InitStatusBar
*
* Status bar helper functions
*
* Yung-Yu Chuang, 2000
*
* History:
* 09/05/2000	Created. (Based on Programming Windows 95) 
****************************************************************************/

HWND InitStatusBar(HWND hwndParent)
{
	HWND hwndSB;

	hwndSB = CreateStatusWindow(
				WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS | CCS_BOTTOM, 
				"Ready", 
				hwndParent, 
				2
			 );

	return hwndSB;
}

char ModeToChar(int mode)
{
	switch (mode) {
	case MENU_MOUSE_MODE:
		return 'M';
	case MENU_ARROW_MODE:
		return 'A';
	case MENU_TRACK_MODE:
		return 'D';
	}

	return ' ';
}

VOID ResetTitle()
{
	char buf[128];

	sprintf(buf,"Snoop x%d [%c] %d-bit (%dx%d)", gnZoom, ModeToChar(gnMode), gnColorBits, gnWindowsWidth, gnWindowsHeight);

	SetWindowText(ghwndApp, buf);
}


/****************************** Screen Scale Header ************************
* Fix Screen Scale Factor Problem. 
* Aside from adding SetProcessDPIAware(), Use ScaleFactor to fix the status bar
*
* Windows Surface pixel:point = 1.5:1
*
* UglyMan.nothinglo, Zorin, 2015
*
* History:
* 04/17/2015	Created. (Based on Programming Windows Surface) 
****************************************************************************/

void initScreenScaleFactor(void){
	// init screen scale factor, pixel:point in windows surface = 1.5:1
	// do this BEFORE SetProcessDPIAware(), else it'll always get 1.
	RECT desktop;
	const HWND hDesktop = GetDesktopWindow();
	GetWindowRect(hDesktop, &desktop);
	GetWindowsDimension(&gnWindowsWidth, &gnWindowsHeight);
	screenScaleFactorY =  gnWindowsHeight / (float)desktop.bottom;
}
